/*
 * $Log: ActiveTCB.java,v $
 * Revision 1.4  2002/01/29 21:41:48  stuart
 * toString()
 *
 * Revision 1.3  1999/08/20  02:05:54  stuart
 * misc routines
 *
 */
package bmsi.edx;

import posix.CPtr;

/** An active TCB in an EDX virtual machine.  This provides shortcuts
    for accessing TCB fields and memory in the same partition.  It is
    a seperate class to keep it Thread safe.  It is not an inner class
    because JDK compiler bugs don't allow blank final variables when
    inner classes are present.
 */

public class ActiveTCB {
  final EDX edx;
  private final CPtr sm;
  private final int[] acore;		// offsets for active partition
  
  public final short actkey;
  public final char acttcb;
  private PROG prog;

  ActiveTCB(EDX edx,CPtr sm,short key,char tcb,int[] acore) {
    this.edx = edx;
    actkey = key;
    acttcb = tcb;
    this.sm = sm;
    this.acore = acore;
    if (acttcb != gtcba(Edxtcb.TCBVER))
      throw new EDXError("Corrupted " + toString());
  }

  private int part(int addr) {
    return acore[addr >> 11] + addr;
  }

  /** Copy part of a byte array into EDX memory. */
  public final void copyIn(int addr,byte[] b,int off,int len) {
    sm.copyIn(part(addr),b,off,len);
  }

  /** Copy EDX memory into a byte array. */
  public final void copyOut(int addr,byte[] b,int off,int len) {
    sm.copyOut(part(addr),b,off,len);
  }

  /** Return a copy in a byte[] of a portion of EDX memory. */
  public final byte[] getBytes(int addr,int len) {
    byte[] b = new byte[len];
    sm.copyOut(part(addr),b,0,len);
    return b;
  }
  /** Copy a byte[] into EDX memory. */
  public final void putBytes(int addr,byte[] b) {
    sm.copyIn(part(addr),b,0,b.length);
  }

  /** Return an EDX TEXT object as a String. No blanks are stripped. The
      length of the String is taken from the byte at <code>addr - 1</code>.
   */
  public final String getText(int addr) {
    return edx.etoa(actkey,addr,getByte(addr-1)&0xFF);
  }

  /** Set the value of an EDX TEXT object.  The length must be less than
      the byte at <code>addr - 2</code>.  The byte at <code>addr - 1</code>
      is set to the length.  The remaining bytes up to the max length at
      <code>addr - 2</code> are set to EBCDIC blank.
   */
  public final void putText(int addr,String val) {
    int len = val.length();
    int maxlen = getByte(addr-2)&0xff;
    if (len > maxlen)
      throw new IllegalArgumentException("String too long for EDX TEXT");
    putByte(addr-1,len);
    edx.atoe(actkey,addr,maxlen,val);
  }

  /** Return an EBCDIC character field as a String.  No blanks are stripped,
      you can call <code>trim()</code> if desired. */
  public final String etoa(int addr,int len) {
    return edx.etoa(actkey,addr,len);
  }

  /** Set an EBCDIC char field from a String.  The String will be truncated
     or padded with blanks to fit in <code>len</code> EBCDIC bytes.
   */
  public final void atoe(int addr,int len,String s) {
    edx.atoe(actkey,addr,len,s);
  }

  /** Return the DWORD at an EDX address in the active partition. */
  public final int getDword(int addr) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    return sm.getInt(part(addr));
  }

  /** Store a DWORD at an EDX address in the active partition. */
  public final void putDword(int addr,int val) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    sm.setInt(part(addr),val);
  }

  /** Return the unsigned WORD at an EDX address in the active partition. */
  public final char getAddr(int addr) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    return (char)sm.getShort(part(addr));
  }

  /** Set the WORD at an EDX address in the active partition of this TCB. */
  public final void putWord(int addr,int val) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    sm.setShort(part(addr),(short)val);
  }

  /** Set the BYTE at an EDX address in the active partition of this TCB. */
  public final void putByte(int addr,int val) {
    sm.setShort(part(addr),(byte)val);
  }

  /** Return the WORD at an EDX address in the active partition. */
  public final short getWord(int addr) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    return sm.getShort(part(addr));
  }

  /** Return the BYTE at an EDX address in the active partition. */
  public final byte getByte(int addr) { return sm.getByte(part(addr)); }

  /** Wakeup this TCB, which is presumably waiting for this external 
      process.
   */
  public final void ready() throws EDXException {
    edx.send(EDX.NC_READY,actkey,acttcb,(short)0);
  }

  public final void post(int ecb,int code) throws EDXException {
    edx.send(EDX.NC_POST,actkey,(char)ecb,(short)code);
  }

  public final void reset(int ecb) throws EDXException {
    edx.send(EDX.NC_RESET,actkey,(char)ecb,(short)0);
  }

  public final short gtcbw(int offset) { return getWord(acttcb + offset); }

  public final char gtcba(int offset) { return getAddr(acttcb + offset); }

  public final void ptcbw(int offset,int val) {
    putWord(acttcb + offset,val);
  }

  /** Finish an EDX terminal operation.
   */
  final void dqterm(int ccb) throws EDXException {
    edx.send(EDX.NC_DQTERM,actkey,acttcb,(short)ccb);
  }

  /** Return the EDX CCB currently assigned to this TCB. */
  public CCB getCCB() {
    return new CCB(edx,gtcba(Edxtcb.TCBCCB));
  }

  /** Return the EDX Program this TCB is part of. */
  public final PROG getPROG() {
    if (prog == null)
      prog = new PROG(this,gtcba(Edxtcb.TCBPLP));
    return prog;
  }

  public int hashCode() {
    return (actkey << 16) + acttcb ^ edx.hashCode();
  }

  public boolean equals(Object obj) {
    if (!(obj instanceof ActiveTCB)) return false;
    ActiveTCB a = (ActiveTCB)obj;
    return (a.edx == edx && a.actkey == actkey && a.acttcb == acttcb);
  }

  public String toString() {
    return "TCB[" + actkey + '-' + Integer.toHexString(acttcb)
      + "]@" + edx.toString();
  }

}
