/*
 * $Log: EDX.java,v $
 * Revision 1.4  2002/01/30 19:25:58  stuart
 * Convert to CVS.
 *
 * Revision 1.3  2000/06/20 03:55:11  stuart
 * more docs, useful toString(), convert IPCException to EDXException
 *
 * Revision 1.2  1998/11/21  00:44:47  stuart
 * getPartitionMap() and wheres()
 *
 * Revision 1.1  1998/10/27  20:01:16  stuart
 * Initial revision
 *
 */
package bmsi.edx;

import posix.*;
import java.util.Vector;
import java.io.UnsupportedEncodingException;

/** A class to interface to a running EDX emulator.

    @author Stuart D. Gathman
    Copyright (C) 1998 Business Management Systems, Inc.
 */
public class EDX implements Edxsys, Edxtcb, Edxprg {
  /** IPC types for use with posix.IPC.ftok(). */
  static final char
    MEMCHAR = 'c',	// EDX SharedMem
    IOCCHAR = 'd',	// EDX device IO MsgQ
    USRCHAR = 'u',	// User extension MsgQ
    TEMCHAR = 't',	// CCB Semaphores
    TRMCHAR = 'T';	// EDX 4978 SharedMem

  /** EDX nucleus message command codes. */
  static final short
    NC_RESET = 1,	/* reset ECB */
    NC_POST = 2,	/* post ECB */
    NC_DEQ = 3,		/* dequeue QCB */
    NC_ATTACH = 4,	/* attach a task */
    NC_READY = 5,	/* add a task to the ready chain */
    NC_TICK = 6,	/* timer tick (or task switch) */
    NC_PCHECK = 7,	/* program check */
    NC_PSTOP = 8,	/* program stop */
    NC_DUMP = 9,	/* dump memory */
    NC_DQTERM = 10,	/* CCB I/O instruction complete */
    NC_BPBREAK = 11,	/* set break point */
    NC_BPWATCH = 12,	/* set watch */
    NC_BPSTEP = 13,	/* set task step */
    NC_NOTIFY = 14;	/* send SIGUSR1 with next senduser() */

  private SharedMem shm;
  private SharedMem smbms;
  private String vtab;

  public static final int MAXPART = 32;

  /** Offsets of shared memory header fields. */
  private static final int instbreak, partlen, partbase, numpart, compart,
    syslgccb, iocmqid, ccbsid, memid, memoff, vthoff, usrmqid;

  /** Offsets of EDX nucleus message fields. */
  private static final int nc_mtype, nc_cmd, nc_key, nc_addr, nc_code, nc_len;

  private final Malloc nucmsg = new Malloc(nc_len);

  static {
    // compute C struct offsets
    CPtr.Struct s = new CPtr.Struct(0);
    instbreak = s.offsetOf(CPtr.CINT_TYPE);
    partlen = s.offsetOf(CPtr.CSHORT_TYPE,MAXPART);
    partbase = s.offsetOf(CPtr.CSHORT_TYPE,MAXPART);
    numpart = s.offsetOf(CPtr.CSHORT_TYPE);
    compart = s.offsetOf(CPtr.CSHORT_TYPE);
    syslgccb = s.offsetOf(CPtr.CSHORT_TYPE);
    s.offsetOf(CPtr.CINT_TYPE);		// edxnuc pid
    s.offsetOf(CPtr.CBYTE_TYPE,40);	// edxnuc cwd
    s.offsetOf(CPtr.CBYTE_TYPE,40);	// edxnuc edxvtab.sys
    s.offsetOf(CPtr.CBYTE_TYPE,40);	// path to nucleus file
    iocmqid = s.offsetOf(CPtr.CINT_TYPE);
    ccbsid = s.offsetOf(CPtr.CINT_TYPE);
    memid = s.offsetOf(CPtr.CINT_TYPE,MAXPART);
    memoff = s.offsetOf(CPtr.CINT_TYPE,MAXPART);
    vthoff = s.offsetOf(CPtr.CINT_TYPE);
    usrmqid = s.offsetOf(CPtr.CINT_TYPE);

    CPtr.Struct m = new CPtr.Struct(0);
    nc_mtype = m.offsetOf(CPtr.CLONG_TYPE);
    nc_cmd = m.offsetOf(CPtr.CSHORT_TYPE);
    nc_key = m.offsetOf(CPtr.CSHORT_TYPE);
    nc_addr = m.offsetOf(CPtr.CSHORT_TYPE);
    nc_code = m.offsetOf(CPtr.CSHORT_TYPE);
    nc_len = m.size();
  }

  private CPtr sm;
  private int core[][];	// offsets in sm by partition and page
  private MsgQ iocq = null;
  /** USERCALL message queue id */
  MsgQ usrq = null;

  private int[][] coreinit() {
    int npart = sm.getCShort(numpart);
    int[][] core = new int[npart][];
    int cpart = sm.getCShort(compart);
    int coff = sm.getCInt(memoff,cpart);
    for (int key = 0; key < npart; ++key) {
      int npage = sm.getCShort(partbase,key) >> 11;
      int plen = sm.getCShort(partlen,key);
      int page;
      core[key] = new int[plen];
      for (page = 0; page < npage; ++page)
	core[key][page] = coff;
      int moff = sm.getCInt(memoff,key);
      /* FIXME: check memid in case of multiple shared mem segments.
         This is highly unlikely on a 32-bit or larger machine. */
      for (; page < plen; ++page)
	core[key][page] = moff;
    }
    iocq = new MsgQ(sm.getCInt(iocmqid));
    usrq = new MsgQ(sm.getCInt(usrmqid));
    return core;
  }

  /** Return the number of partitions in the EDX emulator. */

  public int getNumpart() { return core.length; }

  /** Return the volume table file that defines this EDX VM. */
  public String getVtab() { return vtab; }

  /** Create an interface to an EDX emulator.  We can interface to multiple 
      emulators distinguished by which edxvtab.sys file they are using.

      @param vtab	The edxvtab.sys file that defines which emulator.
   */
  public EDX(String vtab) throws EDXException {
    this.vtab = vtab;
    try {
      shm = new SharedMem(vtab,MEMCHAR);
      sm = shm.attach();
      core = coreinit();
    }
    catch (IPCException e) {
      throw new EDXException("EDX shared memory: " + e.getMessage());
    }
    try {
      smbms = new SharedMem(vtab,TRMCHAR);
    }
    catch (IPCException e) { }
  }

  /** Create an interface to the default EDX emulator.  It is defined
    by the <code>edxvtab.sys</code> file in the directory named by the
    <code>bmsi.edxdir</code> system property, <code>/edx</code> by default.
   */
  public EDX() throws EDXException {
    this(System.getProperty("bmsi.edxdir","/edx") + "/edxvtab.sys");
  }

  CPtr getScreen() {
    try {
      if (smbms != null)
	return smbms.attach(); 
    }
    catch (IPCException e) { }
    return null;
  }

  private int part(int key,int addr) {
    return core[key][addr >> 11] + addr;
  }

  /** Return the BYTE at an EDX address. */
  public final byte getByte(int key,int addr) {
    return sm.getByte(part(key,addr));
  }

  /** Set the BYTE at an EDX address. */
  public final void putByte(int key,int addr,int val) {
    sm.setByte(part(key,addr),(byte)val);
  }

  /** Return the WORD at an EDX address. */
  public final short getWord(int key,int addr) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    return sm.getShort(part(key,addr));
  }

  /** Return the unsigned WORD at an EDX address. */
  public final char getAddr(int key,int addr) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    return (char)sm.getShort(part(key,addr));
  }

  public final void putWord(int key,int addr,int val) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    sm.setShort(part(key,addr),(short)val);
  }

  /** Return the DWORD at an EDX address. */
  public final int getDword(int key,int addr) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    return sm.getInt(part(key,addr));
  }

  /** Store a DWORD at an EDX address. */
  public final void putDword(int key,int addr,int val) {
    if ((addr & 1) != 0) throw new OddByteException(addr);
    sm.setInt(part(key,addr),val);
  }

  private final String noebcdic =
    "Cp037 encoding not installed - need i18n.jar?";

  /** Return an EBCDIC char field as a String. No blanks are stripped,
   you can call <code>trim()</code> if desired. */
  public String etoa(int key,int addr,int len) {
    try {
      // FIXME: use FType
      return new String(getBytes(key,addr,len),"Cp037");
    }
    catch (UnsupportedEncodingException e) {
      throw new RuntimeException(noebcdic);
    }
  }

  /** Set an EBCDIC char field from a String.  The String will be truncated
     or padded with blanks to fit in <code>len</code> EBCDIC bytes.
   */
  public void atoe(int key,int addr,int len,String txt) {
    try {
      // FIXME: use FType
      byte[] b = txt.getBytes("Cp037");
      if (b.length < len) {
        byte[] nb = new byte[len];
	System.arraycopy(b,0,nb,0,b.length);
	for (int i = b.length; i < len; ++i)
	  nb[i] = 0x40;
	b = nb;
      }
      sm.copyIn(part(key,addr),b,0,len);
    }
    catch (UnsupportedEncodingException e) {
      throw new RuntimeException(noebcdic);
    }
  }

  /** Return an EDX TEXT object as a String. No blanks are stripped. */
  public final String getText(int key,int addr) {
    return etoa(key,addr,getByte(key,addr-1)&0xFF);
  }

  /** Return a copy in a byte[] of a portion of EDX memory. */
  public final byte[] getBytes(int key,int addr,int len) {
    byte[] b = new byte[len];
    sm.copyOut(part(key,addr),b,0,len);
    return b;
  }

  /** Copy a byte[] into EDX memory. */
  public final void putBytes(int key,int addr,byte[] b) {
    sm.copyIn(part(key,addr),b,0,b.length);
  }

  /** Copy EDX memory into a byte array. */
  public final void copyOut(int key,int addr,byte[] b,int off,int len) {
    sm.copyOut(part(key,addr),b,off,len);
  }

  /** Copy part of a byte array into EDX memory. */
  public final void copyIn(int key,int addr,byte[] b,int off,int len) {
    sm.copyIn(part(key,addr),b,off,len);
  }


  /** Release the resources used by this EDX interface. Subsequent uses
      will throw an exception. */
  public void dispose() {
    shm.dispose();
    nucmsg.free();
    if (smbms != null)
      smbms.dispose();
  }

  synchronized void send(short cmd,short key,char adr,short code)
    throws EDXException {
    nucmsg.setCInt(nc_mtype,IPC.pid);
    nucmsg.setCShort(nc_cmd,cmd);
    nucmsg.setCShort(nc_key,key);
    nucmsg.setCShort(nc_addr,(short)adr);
    nucmsg.setCShort(nc_code,code);
    try {
      iocq.send(nucmsg,nc_len,0);
    }
    catch (IPCException e) {
      throw new EDXException(e.getMessage());
    }
    sm.setCInt(instbreak,1);
  }

  /** Wakeup an EDX task. */
  public final void ready(int key,int tcb) throws EDXException {
    send(NC_READY,(short)key,(char)tcb,(short)0); }

  /** Deq an EDX QCB. */
  public final void deq(short key,char qcb,short code) throws EDXException {
    send(NC_DEQ,key,qcb,code); }

  /** Post an EDX event. */
  public final void post(short key,char ecb,short c) throws EDXException {
    send(NC_POST,key,ecb,c); }

  /** Attach an EDX task. */
  public final void attach(short k,char tcb,short c) throws EDXException {
    send(NC_ATTACH,k,tcb,c); }

  /** Cause a program check in a running EDX program. */
  public final void pcheck(short k,char a,short c) throws EDXException {
    send(NC_PCHECK,k,a,c); }

  /** Cancel an EDX program. This has the same effect as if the EDX program
      were to execute PROGSTOP as the next intruction.
      @param k the address key of the program
      @param a the address of the program
      @param c the PROGSTOP code
   */
  public final void cancel(short k,char a,short c) throws EDXException {
    send(NC_PSTOP,k,a,c); }

  /** Interrupt EDX execution so that the emulator can check timer queues,
    run a higher priority TASK, etc. */
  public final void tick() throws EDXException {
    send(NC_TICK,(short)0,(char)0,(short)0); }

  /** Start $INITIAL in a newly loaded edxnuc.  Does nothing if $INITIAL
      has already been run or the edxnuc is corrupted.
   */
  public synchronized void startinitial() throws EDXException {
    if (sm == null) return;

    char inittcb = getAddr(0, INITTCB);	// get initial TCB address from CVT
    putWord(0,INITTCB,0);		// zero it out
    if (inittcb == 0 || (inittcb & 1) != 0
      || getAddr(0, inittcb + TCBVER) != inittcb)
      return;
    putWord(0,inittcb + TCBPLP,0);
    putWord(0,inittcb + TCBCCB,sm.getCShort(syslgccb));
    putWord(0,inittcb + TCBADS,0);
    attach((short)0, inittcb, (short)-1);
  }

  /** Get an active TCB context.  This determines the default partition as well
    for a short form of the get/put methods.
    @param key	the address key of the tcb
    @param tcb	the tcb address or 0
    @return a TCB context
   */
  public ActiveTCB active(int key,int tcb) {
    return new ActiveTCB(this,
      sm,(short)key,(char)tcb,core[key]);
  }

  /** Return a snapshot of the memory map for an EDX partition.
    @param key	the address key for the partition. Partition 1 is 
    	address key 0.
    @return an array of map entries for the partition, or null
    	if the partition is not allocated.
    @exception ArrayIndexOutOfBoundsException if the address key
    	is not between 0 and <code>getNumpart()</code>.
   */
  public MapEntry[] getPartitionMap(int key) {
    if (key < 0 || key >= core.length)
      throw new ArrayIndexOutOfBoundsException("key="+key);
    char mem = getAddr(0,MEMLIS);
    char map = getAddr(0,mem + key * 2);
    char entaddr = getAddr(0,map);
    char entsize = getAddr(0,map + 2);
    if (entaddr == 0xFFFF || entaddr == 0 && entsize == 0)
      return null;	// partition not present
    final Vector v = new Vector();
    while (entaddr != 0xFFFF) {
      MapEntry m;
      switch (entsize >>> 8) {
      case 0x80:
	m = new MapEntry(MapEntry.PROG,entaddr,(short)(entsize & 0x00FF));
	break;
      case 0x00:
      case 0x01:
	m = new MapEntry(MapEntry.FREE,(char)(entaddr << 8),(short)entsize);
	break;
      default:
	m = new MapEntry(MapEntry.DATA,entaddr,(short)(entsize & 0x00ff));
      }
      v.addElement(m);
      map += 4;
      entaddr = getAddr(0,map);
      entsize = getAddr(0,map + 2);
    }
    MapEntry[] a = new MapEntry[v.size()];
    v.copyInto(a);
    return a;
  }

  /** Find a running EDX program.
    @param progname	the name of the program
    @return the first program found with a matching name, or null
    		if not found
   */
  public PROG wheres(String progname) {
    for (int k = 0; k < core.length; ++k) {
      MapEntry[] map = getPartitionMap(k);
      for (int i = 0; i < map.length; ++i) {
	MapEntry m = map[i];
	if (m.type == MapEntry.PROG) {
	  String name = getText(k,m.ldpt + PRGNAM).trim();
	  if (name.equals(progname)) {
	    char tcb = getAddr(k,m.ldpt + PRGTCB);
	    return new PROG(active(k,tcb),m.ldpt);
	  }
	}
      }
    }
    return null;
  }

  public String toString() {
    return "EDX[" + vtab  + ']';
  }

}
