/*
 * $Log: EDX4978.java,v $
 * Revision 1.5  2002/01/29 21:44:05  stuart
 * map NUL to Unicode blank
 *
 * Revision 1.4  2001/03/23 19:30:47  stuart
 * support interrupting a waitkey programmatically
 *
 * Revision 1.3  2000/11/17 21:28:51  stuart
 * EDXTerm Support
 *
 * Revision 1.2  1999/08/20 02:11:15  stuart
 * fix graphics mapping
 * use EdxjavaRequest.termctrl
 * some convenience routines
 *
 */
package bmsi.edx;

import posix.*;

/** Access a static 4978 EDX screen from Java.
    @author Stuart D. Gathman
    Copyright (C) 1999 Business Management Systems, Inc.
 */
public class EDX4978 extends CCB {
  static final byte	// used by EDXTerm
    DEVUPDATE =	1,
    DEVWAITKEY= 2,
    DEVBELL   =	3,
    DEVCCALL  =	4;
  protected static final int MAXSIZE = 1920;
  private static final int update;	// screen needs updating
  private static final int wakeup;	// terminal process needs wakeup signal
  private static final int first, last;	// screen area not in sync
  private static final int ccbver;	// verify CCB association
  private static final int pos, op;
  private static final int lsiz, msiz;	// line size, map size
  private static final int org;		// video origin (for scrolling)
  private static final int map;		// screen map

  private int offset;
  private CPtr smbms;
  private ActiveTCB tcb;
  private int row,col;

  private static final int size;

  static {
    CPtr.Struct s = new CPtr.Struct(0);
    update = s.offsetOf(CPtr.CINT_TYPE);
    wakeup = s.offsetOf(CPtr.CINT_TYPE);
    first = s.offsetOf(CPtr.CINT_TYPE);
    last = s.offsetOf(CPtr.CINT_TYPE);
    ccbver = s.offsetOf(CPtr.CSHORT_TYPE);
    pos = s.offsetOf(CPtr.CSHORT_TYPE);
    op = s.offsetOf(CPtr.CSHORT_TYPE);
    lsiz = s.offsetOf(CPtr.CSHORT_TYPE);
    msiz = s.offsetOf(CPtr.CSHORT_TYPE);
    org = s.offsetOf(CPtr.CSHORT_TYPE);
    map = s.offsetOf(CPtr.CBYTE_TYPE,MAXSIZE);
    size = s.size();
  }

  private EdxjavaRequest req;

  EDX4978(ActiveTCB tcb) {
    this(tcb.edx,tcb.gtcba(Edxtcb.TCBCCB));
    this.tcb = tcb;
  }

  EDX4978(EDX edx,char ccba) {	// used by Edxterm
    super(edx,ccba);
    offset = gccbw(CCBVTID) * size;
    smbms = edx.getScreen();
    if (smbms == null)
      throw new EDXError("EDX4978 shared screen buffer missing");
    if ((char)getShort(ccbver) != ccba)
      throw new EDXError("EDX4978 shared screen buffer corrupted");
    Thread t = Thread.currentThread();
    if (t instanceof EdxjavaRequest)
      req = (EdxjavaRequest)t;
  }

  // used by EDXTerm
  protected void setUpdate(int v) { smbms.setCInt(update + offset,v); }
  protected int getUpdate() { return smbms.getCInt(update + offset); }

  void s1sync() {
    while (getUpdate() != 0) {
      int ifirst = offset + first;
      int ilast = offset + last;
      int beg = smbms.getCInt(ifirst);
      int end = smbms.getCInt(ilast);
      smbms.setCInt(ifirst,getShort(msiz));
      smbms.setCInt(ilast,0);
      setUpdate(0);
      int nbeg = smbms.getCInt(ifirst);
      int nend = smbms.getCInt(ilast);
      if (beg > nbeg) beg = nbeg;
      if (end < nend) end = nend;
      update(beg,end);
      if (getUpdate() == 0)
        smbms.setCInt(offset + wakeup,1);
    }
  }

  protected void update(int beg,int end) { }

  private int getShort(int off) {
    return smbms.getCShort(offset + off);
  }

  /** Unicode chars corresponding to the line drawing codes
      on the BMS 4978 client.  The line drawing codes occupy
      the first 31 positions after 0.
   */
  static final char graph[] = {
    0x20, 0x2666, 0x2592, 0x2510, 0xb0, 0x2500, 0x250c, 0x252c,
    0xb1, 0x2518, 0x2502, 0x2524, 0x2514, 0x2534, 0x251c, 0x253c,
    0x2588, 0x2590, 0x2580, 0x2557, 0x258c, 0x2550, 0x2554, 0x2566,
    0x2584, 0x255d, 0x2551, 0x2563, 0x255a, 0x2569, 0x2560, 0x256c
  };

  /** Clear the screen from the current cursor position to the end.
    @param prot Clear to protected spaces if true.
   */
  public void cleareos(boolean prot) {
    int i = getShort(pos);
    byte[] b = new byte[MAXSIZE - i];
    if (prot)
      for (int j = 0; j < b.length; ++j)
        b[j] |= 0x80;
    write(i,b,0,b.length);
  }

  /** Write a string at the current cursor position. 
    @param s	The string to write.
    @param prot	Write protected chars if true.
   */
  public final void write(String s,boolean prot) {
    write(s,s.length(),prot);
  }

  /** Write a string at the current cursor position.  The String will
    be blank padded or truncated to the requested number of columns.
    @param s	The string to write.
    @param len	The size in columns to write.
    @param prot	Write protected chars if true.
   */
  public void write(String s,int len,boolean prot) {
    write(getShort(pos),s,len,prot);
  }

  public void write(int row,int col,String s,int len,boolean prot) {
    write(getPos(row,col),s,len,prot);
  }

  public void write(int pos,String s,int len,boolean prot) {
    byte[] b = new byte[len];
    int mask = prot ? 0x80 : 0;
    int slen = s.length();
    if (slen > len) slen = len;
    int i = 0;
    while (i < slen) {
      int c = s.charAt(i);
      switch (c) {
      case 0x2666: c= 1; break; case 0x2592: c= 2; break;
      case 0x2510: c= 3; break; case 0xb0: c= 4; break;
      case 0x2500: c= 5; break; case 0x250c: c= 6; break;
      case 0x252c: c= 7; break; case 0xb1: c= 8; break;
      case 0x2518: c= 9; break; case 0x2502: c= 10; break;
      case 0x2524: c= 11; break; case 0x2514: c= 12; break;
      case 0x2534: c= 13; break; case 0x251c: c= 14; break;
      case 0x253c: c= 15; break; case 0x2588: c= 16; break;
      case 0x2590: c= 17; break; case 0x2580: c= 18; break;
      case 0x2557: c= 19; break; case 0x258c: c= 20; break;
      case 0x2550: c= 21; break; case 0x2554: c= 22; break;
      case 0x2566: c= 23; break; case 0x2584: c= 24; break;
      case 0x255d: c= 25; break; case 0x2551: c= 26; break;
      case 0x2563: c= 27; break; case 0x255a: c= 28; break;
      case 0x2569: c= 29; break; case 0x2560: c= 30; break;
      case 0x256c: c= 31; break;
      }
      b[i++] = (byte)(c & 0x7F | mask);
    }
    if (i < len) {
      byte c = (byte)(0x20 | mask);
      do
	b[i++] = c;
      while (i < len);
    }
    write(pos,b,0,b.length);
  }

  /** Read the contents of an unprotected field as a String. */
  public String input(int row,int col) {
    int cnt = 80 - col;
    byte[] b = new byte[cnt];
    int i = getPos(row,col);
    read(i,b,0,cnt);
    return input(b);
  }

  public static String input(byte[] b) {
    int i;
    int cnt = b.length;
    for (i = 0; i < cnt; ++i)
      if (b[i] < 0) break;
    cnt = i;
    char[] c = new char[cnt];
    for (i = 0; i < cnt; ++i) {
      char cc = (char)b[i];
      c[i] = (cc < graph.length) ? graph[cc] : cc;
    }
    return new String(c);
  }

  /** Unprotect chars on the screen beginning at the current position.
    @param cnt The number of chars to unprotect.
   */
  public void unprotect(int cnt) {
    byte[] b = new byte[cnt];
    int off = getShort(pos);
    read(off,b,0,cnt);
    for (int i = 0; i < cnt; ++i)
      b[i] &= 0x7F;
    write(off,b,0,cnt);
  }

  /** Return the current row. */
  public final int getRow() { return row; }

  /** Return the current column. */
  public final int getCol() { return col; }

  /** Set the cursor position.
      @param row	The row offset (0 origin)
      @param col	The column offset (0 origin)
   */
  public void setCursor(int row,int col) {
    //System.err.println("lsiz="+getShort(lsiz));
    this.row = row;
    this.col = col;
    smbms.setCShort(offset + pos,(short)getPos(row,col));
  }

  /** Return the current offset within the raw screen buffer.  This
    can be set with <code>setCursor()</code> or by the user
    during a <code>waitkey()</code>.
   */
  public int getPos() { return getShort(pos); }

  /** Return the raw screen offset for a row and column. */
  public int getPos(int row, int col) {
    return (getShort(lsiz) * row + col) % getShort(msiz);
  }

  /** Read raw bytes from the screen. Each byte is a 7 bit ASCII char.
      The high order bit is set for protected chars.  
    @param pos The offset within the raw screen buffer.
    @param buf A buffer to hold the raw bytes read.
    @param off The offset within buf to store the raw bytes at.
    @param cnt The number of bytes to read.
   */
  public void read(int pos,byte[] buf,int off,int cnt) {
    smbms.copyOut(pos + map + offset,buf,off,cnt);
  }

  /** Write raw bytes to the screen.  Each byte is a 7 bit ASCII char.
      The high order bit is set for protected chars.  
    @param pos The offset within the raw screen buffer.
    @param buf The raw bytes to write.
    @param off The offset within buf to write from.
    @param cnt The number of bytes to write.
   */
  public void write(int pos,byte[] buf,int off,int cnt) {
    pos %= getShort(msiz);
    smbms.copyIn(pos + map + offset,buf,off,cnt);
    int i = first + offset;
    if (pos < smbms.getCInt(i))
      smbms.setCInt(i,pos);
    i = last + offset;
    if (pos + cnt > smbms.getCInt(i))
      smbms.setCInt(i,pos + cnt);
    setUpdate(1);
  }

  /** Flush buffered changed to the 4978 screen. */
  public void sync() {
    pccbb(CCBNXTL,row);
    pccbb(CCBCPOS,col);
    termctrl(0x1c30);
  }

  /** Wait for the user to press a PFKEY.
      @return the PFKEY index.
   */
  public int waitkey() {
    return waitkey(false);
  }

  private boolean inWaitkey;
  /** Wait for the user to press a PFKEY.  If tab mode is selected, a
    pseudo PFKEY of 100 is returned when the user leaves the current field.
    @param tabs true to select tab mode.
   */
  public synchronized int waitkey(boolean tabs) {
    pccbb(CCBNXTL,row);
    pccbb(CCBCPOS,col);
    inWaitkey = true;
    termctrl(tabs ? 0xA030: 0x2030);
    inWaitkey = false;
    int i = getShort(pos);
    int l = getShort(lsiz);
    row = i / l;
    col = i % l;
    return tcb.gtcbw(Edxtcb.TCBCO2);
  }

  /** Detach this screen object from EDX.  Further attempts to use
      this instance will result in NullPointerException. */
  void dispose() {
    req = null;
    smbms = null;
    tcb = null;
  }

  private void termctrl(int cmd) {
    req.termctrl(cmd);
  }

  /** Post the CCB event as if the user has pressed a function key.
    The pfkey parameter does not work.  The application always ends
    up with a PFKEY value of -1.  I do not know why.
   */
  void postkey(int k) throws EDXException {
    if (ownedBy(tcb) && inWaitkey) {
      // FIXME: race condition if user presses an FKEY an instant
      // before we do this...
      smbms.setCShort(op,(short)k);
      edx.send(EDX.NC_DQTERM,tcb.actkey,tcb.acttcb,(short)ccba);
    }
  }

}
