/*
 * $Log: EDXPrintJob.java,v $
 * Revision 1.4  2002/01/29 21:39:14  stuart
 * Send 17CPI init for wide reports.
 *
 * Revision 1.3  2001/03/20  16:35:32  stuart
 * Interface with CUFAX (send report as fax or email)
 *
 * Revision 1.2  2001/03/15 20:37:45  stuart
 * Implement abort "dialog" (PRESS ENTER TO INTERRUPT PRINTING)
 *
 * Revision 1.1  2000/11/17 19:34:08  stuart
 * Initial revision
 *
 */
package bmsi.edx;

import bmsi.db.DataBuffer;
import bmsi.db.StringField;
import bmsi.db.IntegerField;
import bmsi.edx.fs.FSPager;
import bmsi.edx.fs.FSPanel;
import bmsi.util.TerminatedException;
import java.io.IOException;
import java.io.OutputStream;
import java.io.FilterOutputStream;
import java.io.PrintWriter;

/** Traditional BMS user interface elements for printing in the EDX
    environment.
<pre>
	EDXPrintJob pjob = new EDXPrintJob(req,"Aged Receivables",132);
	PrintWriter wtr = new PrintWriter(pjob.getOutputStream(),true);
</pre>
    @author Stuart D. Gathman
    Copyright (C) 2000 Business Management Systems, Inc.
 */
public class EDXPrintJob {
  private EDXServletRequest req;
  private String pname;
  private int linesize;
  private int pagesize = 66;
  private String title;
  private static final SelectPrinter selprtr = new SelectPrinter();
  private OutputStream out;

  /** Create an EDXPrintjob.
    @param req	The EDXServletRequest associated with this printjob
    @param title	The title of the report
    @param linesize	The width of the report
    */
  public EDXPrintJob(EDXServletRequest req,String title,int linesize) {
    this.req = req;
    this.title = title;
    this.linesize = linesize;
  }

  /** Select the printer for the job using the traditional BMS interface. */
  public void selprtr() throws IOException {
    byte[] parm = new byte[104];
    DataBuffer buf = new DataBuffer(parm);
    buf.setEncoding("Cp037");
    StringField ptr = new StringField(buf,0,8);
    IntegerField xline = new IntegerField(buf,20,2);
    IntegerField xtctl = new IntegerField(buf,22,2);
    StringField xtitle = new StringField(buf,24,80);
    xline.value(80);		// Don't include screen
    xtitle.value(title);
    xtctl.value(0x5050);
    try {
      req.shell("SELPRTR,MSLIB",parm);
      linesize = xline.value();
      pname = ptr.value().trim();
      pagesize = 66;
    }
    catch (EDXShellException x) {
      x.printStackTrace();
      throw new IOException("CANNOT LOAD SELPRTR,MSLIB");
    }
  }

  /** Pagesize of printer. */
  public int getPagesize() { return pagesize; }
  /** EDX Printer name.  This is the 8 character name as used by edxnuc.  */
  public String getPrinter() { return pname; }
  /** Set the EDX printer to use.  If the printer does not exist, then
    getOutputStream() will go to the screen.
   */
  public void setPrinter(String name) { pname = name; }
  /** Report Title. */
  public String getTitle() { return title; }
  /** Line size of printer. */
  public int getLinesize() { return linesize; }

  private final byte[] faxbuf = new byte[70];
  private final DataBuffer faxparms = new DataBuffer(faxbuf);
  { faxparms.setEncoding("Cp037"); }
  public boolean abortDialog = false;
  /** PRINTER NAME */
  private final StringField fxprtr = new StringField(faxparms,0,8);
  /** DEFAULT CUSTOMER #/VENDOR # */
  private final IntegerField fxcuno = new IntegerField(faxparms,8,4);
  /** PARMFLG (0=CUST,1=VEND,2=AGENT) */
  private final IntegerField fxflag = new IntegerField(faxparms,12,2);
  /** FAX LOG TAG */
  private final StringField fxtag = new StringField(faxparms,14,15);
  /** DEFAULT CONTACT NAME */
  private final StringField fxdefnm = new StringField(faxparms,29,32);
  /** PARM1 (3 BYTES USED FOR AGENT) */
  private final StringField fxparm1 = new StringField(faxparms,61,9);
  //* DEFAULT CONTACT FAX NUMBER */
  //private final StringField fxdeffx = new StringField(faxparms,70,16);

  static class CB implements EDXServlet {
    PrintWriter out;
    public void init() { }
    public void destroy() { }

    public short callJava(EDXServletRequest req) {
      byte[] b = req.getParm();
      try {
	String s = new String(b,"Cp037").trim();
	out.println("\\\\"+s);
      }
      catch (IOException x) {
	x.printStackTrace();
	return 1;
      }
      return 0;
    }
  }

  public synchronized OutputStream getOutputStream() throws IOException {
    if (out != null) return out;
    if (pname == null) selprtr();
    EDX4978 sp = req.getScreen();
    synchronized (selprtr) {
      selprtr.setTermEnv("EDXUID",String.valueOf(sp.getUser()));
      selprtr.setTermEnv("EDXGID",String.valueOf(sp.getGroup()));
      selprtr.setTermEnv("EDXTERM",sp.getName());
      selprtr.setTermEnv("PATH",
        "/bin:/usr/bin:/usr/local/bin:/bms/bin:/bms/bin/edx");
      out = selprtr.getOutputStream(pname);
      byte[] init;
      if (linesize > 80)
        init = selprtr.getEscape(pname,"17CPI");
      else
        init = selprtr.getEscape(pname,"10CPI");
      if (init != null)
        out.write(init);
    }
    if (out != null && pname.startsWith("FAX")) {
      fxtag.value(req.getTCB().getPROG().getName());
      Edxjava j = req.getEdxjava();
      int handle = j.addEntry("bmsi.edx.EDXPrintJob$CB");
      try {
	fxprtr.value("  " + String.valueOf(handle));
        ((CB)j.getEDXServlet(handle)).out = new PrintWriter(out,true);
	req.shell("CUFAX,BTGL",faxbuf);
      }
      catch (EDXShellException x) {
	x.printStackTrace();
	throw new IOException("CANNOT LOAD CUFAX,BTGL");
      }
      finally {
        j.removeEntry(handle);
      }
      if (!pname.equals(fxprtr.value()))
        System.err.println("CUFAX changed printer "
	  + pname + " -> " + fxprtr.value());
        //throw new IOException("CUFAX changed printer name.");
    }
    if (out == null) {
      out = new FSPager(sp);
      pagesize = 23;
    }
    else if (abortDialog) {	// setup abort dialog
      //System.err.println("Abort Dialog Setup");
      final EDXAbortDialog adlg =
	new EDXAbortDialog(sp,"USING PRINTER: " + pname);
      out = new FilterOutputStream(out) {
	public void write(byte[] buf,int pos,int len) throws IOException {
	  if (adlg.checkAbort())
	    throw new TerminatedException("Report terminated by user");
	  super.write(buf,pos,len);
	}
	public void close() throws IOException {
	  super.close();
	  adlg.dispose();
	}
      };
      adlg.start();
    }
    return out;
  }
}

class EDXAbortDialog implements Runnable {
  private EDX4978 sp;
  private FSPanel pan = new FSPanel();
  private boolean aborted = false;
  private boolean paused = false;
  private String msg;
  private Thread thread = new Thread(this,"EDXAbort");

  public EDXAbortDialog(EDX4978 sp,String msg) {
    this.sp = sp;
    this.msg = msg;
  }

  public void start() {
    thread.start();
  }

  public void run() {
    System.err.println("Abort Dialog Startup");
    pan.paint(sp);
    pan.centerLine(8,msg);
    try {
      do {
	setPaused(false);
	pan.centerLine(10,"PRESS ENTER TO INTERRUPT PRINTING");
	pan.centerLine(11,"");
	int pfkey = sp.waitkey();
	if (pfkey < 0) {
	  pan.centerLine(10,"PRESS ENTER TO FINISH");
	  sp.waitkey();
	  break;
	}
	if (pfkey > 0)
	  System.err.println("pfkey = " + pfkey);
	setPaused(true);
	pan.centerLine(10,"PRESS ENTER TO CONTINUE PRINTING");
	pan.centerLine(11,"PRESS F1 TO ABORT PRINTING");
      } while (sp.waitkey() == 0);
    }
    catch (EDXError x) {
      System.err.println(x.toString());
    }	// interrupted by dispose
    finally {
      abort();
    }
    System.err.println("Abort Dialog Shutdown");
  }

  public synchronized void setPaused(boolean f) {
    if (f != paused && !aborted) {
      paused = f;
      notifyAll();
    }
  }

  private synchronized void abort() {
    if (!aborted) {
      paused = false;
      aborted = true;
      notifyAll();
    }
  }

  public synchronized boolean checkAbort() {
    try { while (paused) wait(); }
    catch (InterruptedException x) { }
    return aborted;
  }


  /** Monitored process is complete, tell abort thread to go away. */
  public synchronized void dispose() {
    try {
      while (!aborted) {
        try { sp.postkey(50); }
	catch (EDXException e) { thread.interrupt(); }
	wait();
      }
    }
    catch (InterruptedException x) { }
  }
}
