/*
 * $Log: EDXServlet.java,v $
 * Revision 1.2  2002/01/29 21:38:26  stuart
 * Expand doc comments.
 *
 * Revision 1.1  1998/10/31  22:14:02  stuart
 * Initial revision
 *
 */
package bmsi.edx;

/** Java classes that wish to be called from EDX via the CALL JAVA interface
  must implement this interface.  The EDX program must first obtain an
  id for the Java class, then use the id to make calls.
<pre>
MYID	 DATA F'0'
AESSHIP	 TEXT 'bmsi.ox.AESShip'
FILNO	 DATA D'0',CL2' '
	 ...
	 CALL FINDJAVA,MYID,(AESSHIP)	GET THE ID FOR THE AESShip CLASS
	 ...
	 CALL JAVA,MYID,(FILNO),6	GENERATE AN EDI MESSAGE
</pre>
  
  The <code>FINDJAVA</code> routine creates a new instance of the Java class
  only if it has not already been assigned an id.
  The id assigned to a Java class can be removed with 

<pre>
	 CALL UNBIND,MYID
</pre>

  In some cases, multiple instances of a Java class may be desired.  To
  unconditionally load a new instance:

<pre>
	 CALL LOADJAVA,ID,(CLASS$)
</pre>

  which creates a new instance.  Be sure to call <code>UNBIND</code> for 
  ids created with LOADJAVA.
<p>
  <code>FINDJAVA</code> and <code>LOADJAVA</code> will return a negative
  id if the Java class is not found or cannot be loaded or does not
  implement EDXServlet.  The <code>JAVA</code> call will post error
  code 701 if the id passed is negative.  This arrangement lets you
  test whether a class is present.  But if you don't check, an EDL
  exception will be thrown when you try to call it.  
<p>
  You may want to call a Java class only if present, for example, only
  if Java is available at a customer site.  In this case, test whether
  the FINDJAVA succeded like this:
<pre>
	IF ID,GE,0
	  CALL JAVA,ID,(PARMS),LEN
	ENDIF
</pre>
  @author Stuart D. Gathman
 */
public interface EDXServlet {

  /** Called when the servlet is loaded.  */
  public void init();

  /** Called by the EDX program.  This method can be called from multiple
      threads.  If an application is not designed to handle parallel
      execution, use the <code>synchronized</code> keyword when
      implementing this method to serialize calls.  Any exception
      thrown from callJava results in CALL ERRPOST,703 on the EDL side.
      @return a word to put in <code>$TCBCO2</code> of the calling EDX task.
      @see Errpost
   */
  public short callJava(EDXServletRequest req) throws Exception;

  /** Called when the servlet is removed.  This method will also be
      called when the Java VM shuts down because the EDX VM has terminated.
      After <code>destroy</code> returns, any calls to <code>callJava</code>
      that are still in progress will be terminated abrubtly. If this is
      not acceptable, synchronizing both <code>callJava</code> and
      <code>destroy</code> will ensure that <code>callJava</code> will 
      complete before <code>destroy</code> is entered.
   */
  public void destroy();

  /** EDXServlets can throw this from callJava() to call ERRPOST
      on the EDL side.  BMS programmers, be sure to document any error
      codes you ERRPOST in /bms/bin/perror.list and in ERROR.DEF.
   */
  public static class Errpost extends Exception {
    public final short code;
    public Errpost(int code,String msg) {
      super(msg);
      short c = (short)code;
      this.code = (c > 0) ? c : 705;
    }
  }
}
