/*
 * $Log: EDXServletRequest.java,v $
 * Revision 1.4  2002/01/30 19:25:58  stuart
 * Convert to CVS.
 *
 * Revision 1.3  2002/01/29 21:37:46  stuart
 * Expand doc comments.
 *
 * Revision 1.2  1999/08/11  03:23:56  stuart
 * EDL callback and SHELL
 *
 * Revision 1.1  1998/10/31  21:17:56  stuart
 * Initial revision
 *
 */
package bmsi.edx;

import java.io.OutputStream;
import java.io.InputStream;
import java.io.IOException;

/** This interface gives access to the EDX properties associated 
    with the EDX program making a CALL JAVA request.

   @author Stuart D. Gathman
 */

public interface EDXServletRequest {

  /** Return an OutputStream connected to the EDX4978 or EDX4974 emulator
     for the EDX terminal attached to this Request.  
     The first time this is called for a
     Request, the EDX4978 will clear the screen and enter
     "roll screen" mode.  The EDX4978 will return to the EDX screen
     when this Request is disposed.
   */
  public OutputStream getOutputStream() throws IOException;

  /** Return an InputStream connected to the EDX4978 or EDX4974 emulator
      for the EDX terminal attached to this Request.
     The first time this is called for a
     Request, the EDX4978 will clear the screen and enter
     "roll screen" mode.  The EDX4978 will return to the EDX screen
     when this Request is disposed.
   */
  public InputStream getInputStream() throws IOException;

  /** Retrieve the parameter block passed by the EDX program.  This
      is a simple array of bytes, and interpretation is application
      specific.  The <code>bmsi.xopen.FType</code> class is helpful
      for converting EDX data to Java format.  Hint: Java uses ISO
      names for 8-bit encodings - the ISO name for EBCDIC is Cp037.
      Use the Cp037 encoding to convert EBCDIC chars to unicode using
      either FType methods or String contructors.

      @see bmsi.xopen.FType
   */
  public byte[] getParm();

  /** Return parameter data to the EDX caller.  The return data
      overwrites the block passed as input to the call for as
      many bytes as are provided - but only up to the original size
      of the block.  Return data beyond the block size specified by
      the EDX caller is discarded.  The modified parameter block that
      results can be inspected by calling <code>getParm</code> again.
   */
  public void setParm(byte[] b);

  /** Set the number of parameter bytes return to the EDX caller.
    Return data beyond the block size specified by the EDX caller
    is discarded.
   */
  public void setParmLen(int b);

  /** Return the EDX Task of the caller.  The ActiveTCB object make available
      additonal information such as the PROG and CCB.
   */
  public ActiveTCB getTCB();

  /** Return the default BTAS directory for the calling EDX program.
     The default BTAS directory is passed as $PARM1 to BMS overlays.
      @return null if it is not available for some reason
   */
  public String getBtasdir();

  /** Return static screen access for the callilng EDX program.
      The current implementation can support either static screen
      or roll screen (<code>getOutputStream()</code>), but not
      both.  The EDX task must have the 4978 terminal ENQTed.
      @since 1.2
      @return null if not available or not ENQTed
   */
  public EDX4978 getScreen();

  /** Call back to EDL.  Suspends this EDXServletRequest and transfers
     control back to EDL.  The EDL code returns to Java by an 
     unconditional transfer to RETJAVA.  Recursive calls back again to
     a new Java EDXServlet are not supported at present.
     @since 1.2
     @param op	A code which is put into $TCBCO2 to select what to do
     	in the EDL code.
   */
  public void calledx(int op);

  /** Call a BMS overlay using the <code>SHELL</code> system.  The current
    program is suspended and swapped out to disk.  Another overlay
    is run, and the current program is swapped back in and resumed
    when it terminates.
    @since 1.2
    @param prog	the NAME,VOL of the EDX program overlay to call
    @param b	the parameter bytes passed to and returned from the overlay
    @return the number of bytes returned from the called overlay
    @exception XParmException on XPARM overflow
    @exception EDXShellException on SHELL stack overflow,
    	or other EDX tasks active in the EDX overlay program
    @exception EDXError on IPC failure talking to the EDX VM
    @see PROG#parmPut
    @see PROG#parmGet
   */
  public int shell(String prog,byte[] b) throws EDXShellException;

  /** Run a command from the user's EDX4978 interface.  The command is run
    attached to the users terminal and with the users privileges.  The
    calling EDXServletRequest Thread blocks until the command finishes.
    Yes, the programs run using edx4978 had better be trusted.  Eventually,
    edx4978 should be configured with allowed commands - like sudo.
    <p>
    Unix commands can read binary EDX parameter bytes on file descriptor 4
    and write a binary EDX result record to file descriptor 3.

    @param cmd	the command to execute - interpreted by the shell
    @param b	EDX parameters to pass to the command
    @return the number of EDX parameter bytes returned by the called program
   */
  public int ushell(String cmd,byte[] b) throws EDXShellException;

  /** Return the Edxjava instance this request belongs to. */
  public Edxjava getEdxjava();

}
