/* $Log: Edxjava.java,v $
/* Revision 1.9  2002/03/05 03:40:57  stuart
/* Generalized runedl.
/*
/* Revision 1.8  2002/01/30 19:25:58  stuart
/* Convert to CVS.
/*
 * Revision 1.7  2001/04/27 19:15:54  stuart
 * log stub and Edxjava versions
 * redirect System.out to EDX terminal
 * provide ushell() and getRequest()
 *
 * Revision 1.6  1999/07/30 21:23:14  stuart
 * Support JRESUME for EDL callouts
 *
 * Revision 1.5  1999/07/30  15:15:39  stuart
 * improve error handling, name constants
 *
 * Revision 1.4  1999/04/29  01:57:01  stuart
 * Use HandleTable
 * More robust exception handling
 *
 * Revision 1.3  1998/11/21  00:46:24  stuart
 * Support stop and restart commands
 *
 * Revision 1.2  1998/10/31  20:52:51  stuart
 * Support additional properties of EDXServletRequest.
 * Query the EDL JAVA stub to translate BTCB's to pathnames.
 *
 */
package bmsi.edx;

import bmsi.util.HandleTable;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import lava.clib.Stdio;

/** This class implements another level on top of the CALLUSER interface
   provided by Edxuser.  We provide a way for EDX programs to invoke
   specific Java classes that implement the EDXServlet interface.  This
   class is invoked by loading JAVA,MSLIB in the EDX system -
   usually in $INITIAL.

    @see EDXServlet

    @author Stuart D. Gathman
 */
public class Edxjava implements Edxtcb {
  /** Common data block offsets in JAVA stub */
  private static final short
    JAVA_UID	 = 0,		// id of Edxjava in EDXUSER interface
    JAVA_ECB	 = 2,		// ECB to post for commands to stub
    JAVA_BTCB	 = 8,		// BTCB for translating working dirs
    JAVA_VERS	 = 40;		// Version string for JAVA stub
  /** Request codes from edxjava.e */
  static final short
    JFIND	 = -1,		// find and/or load class and assign id
    JUNBIND	 = -2,		// unbind id and release class
    JLOAD	 = -3,		// unconditionally load class and assign id
    JRESUME	 = -4;		// resume servlet after EDL callout
  /** Result codes returned to edxjava.e */
  static final short
    JOK		 = -1,		// call java succeeded
    JTERM	 = -2,		// terminal I/O request
    JWAITKEY	 = -3,		// wait key
    JCALLEDX	 = -4,
    JNOTFOUND	 = 701,		// class not found
    JINTERN	 = 702,		// uncaught Edxjava exception
    JAPPERR	 = 703,		// uncaught application exception
    JBADID	 = 704,		// invalid EDXServlet id
    JEDXERR	 = 705;		// EDXError in Java application
    
  private final HandleTable v = new HandleTable();
  private final Hashtable tbl = new Hashtable();
  private Edxuser u;
  
  /** An Entry in the Hashtable of EDXServlets by name. */
  private static class Entry {
    short id;	// the handle used by EDL code to reference EDXServlet
    EDXServlet servlet;
    Entry(short id,EDXServlet s) { this.id = id; servlet = s; }
  }
  
  /** Add an EDXServlet by class name.  A new instance of the named calss
    is allocated and initialized.  A -1 is returned if the class could
    not be found, or does not implement bmsi.edx.EDXServlet, or the
    handle table is full.
   @param name	the class name of the EDXServlet to create
   @return the new EDX handle, or -1 on error
   */
  public short addEntry(String name) {
    try {
      Class c = Class.forName(name);
      if (!EDXServlet.class.isAssignableFrom(c)) return -1;
      EDXServlet s = (EDXServlet)c.newInstance();
      s.init();
      int id = v.add(s);
      short jid = (short)id;
      if (id != jid) {
	System.err.println("Edxjava: too many handles");
	removeEntry(id);
	return -1;
      }
      tbl.put(name,new Entry(jid,s));
      return jid;
    }
    catch (Exception x) {
      x.printStackTrace();
    }
    return -1;
  }

  /** Return the EDXServet for an EDX handle. */
  EDXServlet getEDXServlet(int id) { return (EDXServlet)v.get(id); }

  /** Remove an EDX handle. The EDXServlet instance is destroyed after removal.
   @param id the EDX handle of an EDXServlet, such as returned from addEntry()
   */
  public void removeEntry(int id) {
    EDXServlet s = (EDXServlet)v.remove(id);
    if (s != null)
      s.destroy();
  }
  
  /** Find and return a shared EDXServlet.  If the named class is not already
     loaded by name, a new instance is added.  Shared EDXServlet instances
     should be reentrant.  They will be called by multiple Threads.  The same
     handle will be shared among all EDX programs in this nucleus.
     @return the EDX handle of the named EDXServlet
   */
  private short idFromName(String name) {
    System.err.println("FINDJAVA: " + name);
    Entry e = (Entry)tbl.get(name);
    if (e != null) return e.id;
    return addEntry(name);
  }

  private Edxjava(Edxuser u) throws EDXException {
    this.u = u;
    System.err.println("$Id: Edxjava.java,v 1.9 2002/03/05 03:40:57 stuart Exp $");
    String stubvers = getStubVersion();
    if (stubvers != null)
    System.err.println(stubvers);
    System.err.println("Edxjava: init " + new java.util.Date());
  }

  /** Each Edxjava instance has a JAVA EDL program with a TCB which can
      serve one callback request at a time for Edxjava.  This is used
      to translate BTCBs to directory pathnames, load EDL programs, etc.
   */
  private ActiveTCB callback = null;
  private Edxuser.UIMsg response = null;

  /** Query the JAVA EDL stub that loaded Edxjava to translate
      a BTCB to a pathname.
   */
  synchronized String getBtasdir(byte[] btcb) {
    ActiveTCB j = u.getActiveTCB();	// the JAVA EDL stub
    if (j == null) return null;
    char prog = j.gtcba(TCBPLP);
    char pext = j.getAddr(prog + Edxprg.PRGEXT);	// proghdr extension
    int bta = pext + 8;
    try {
      while (callback != null)	// callback in use
	wait();
      callback = j;
      try {
	j.copyIn(bta,btcb,0,32);
	j.post(pext + 2,bta);
	while (response == null)
	  wait();
	char parm = response.getParm();
	response.uidone(JOK);
	return j.getText(parm).trim();
      }
      finally {		// make sure we don't hold on to callback
	response = null;
	callback = null;
	notify();
      }
    }
    catch (Exception x) { }
    return null;
  }

  /** Handle a CALLUSER from edxjava.e.  The first parameter word
    is either a command < 0, or a handle >= 0. 
   */
  void calluser(Edxuser.UIMsg m) {
    ActiveTCB edx = m.getActiveTCB();
    if (edx.equals(callback)) {
      synchronized (this) {
	callback = null;
	response = m;
	notifyAll();
      }
      return;
    }
    char parm = m.getParm();
    short id = edx.getWord(parm);
    //System.err.println("parm="+(int)parm+" id="+id+" op="+m.getOpword());
    switch (id) {
    case JFIND:
      id = idFromName(edx.getText(edx.getAddr(parm+2)));
      edx.putWord(parm,id);
      m.uidone ((id < 0) ? JNOTFOUND : JOK);
      return;
    case JUNBIND:
      removeEntry(edx.getWord(parm+2));
      m.uidone(JOK);
      return;
    case JLOAD:
      id = addEntry(edx.getText(edx.getAddr(parm+2)));
      edx.putWord(parm,id);
      m.uidone ((id < 0) ? JNOTFOUND : JOK);
      return;
    case JRESUME:
      EdxjavaRequest.resume(edx,m);
      return;
    default:
      EDXServlet s = (EDXServlet)v.get(id);
      if (s == null) {
	m.uidone(JBADID);
	break;
      }
      EdxjavaRequest r = EdxjavaRequest.getRequest(m,parm);
      r.run(s,this);
    }
  }

  /** Shutdown this CALL JAVA interface. */
  void userquit(boolean abnormal) {
    System.err.println("Edxjava: shutdown");
    Object[] s = v.removeAll();
    tbl.clear();
    for (int i = 0; i < s.length; ++i)
      try {
	((EDXServlet)s[i]).destroy();
      }
      catch (Exception e) {
	e.printStackTrace();
      }
  }

  /** Return the EDXServletRequest corresponding to the current Thread.
    @return null if there is no request associated with the current Thread
   */
  public static EDXServletRequest getRequest() {
    Thread t = Thread.currentThread();
    if (t instanceof EdxjavaRequest)
      return (EdxjavaRequest)t;
    return null;
  }

  public void who() {
    Stdio.printf("%3s %8s %s\n",new Object[] {
      "Hnd", "Term", "Class"
    });
    Enumeration e = tbl.elements();
    while (e.hasMoreElements()) {
      Entry t = (Entry)e.nextElement();
      Stdio.printf("%3d %8s %s\n",new Object[] {
        new Integer(t.id), "??", t.servlet.getClass().getName()
      });
    }
  }

  /** Run a command from the user's EDX4978 interface.  Equivalent
  to <code>getRequest().ushell(cmd,b)</code>, except that no 
  NullPointerException is thrown if there is no Request for the current Thread.
    <p>
    Unix commands can read binary EDX parameter bytes on file descriptor 4
    and write a binary EDX result record to file descriptor 3.

    @param cmd	the command to execute - interpreted by the shell
    @param b	EDX parameters to pass to the command
    @return the number of EDX parameter bytes returned by the called program
    @see EDXServletRequest#ushell
   */
  public static int ushell(String cmd,byte[] b) throws EDXShellException {
    EDXServletRequest req = getRequest();
    if (req == null) return 0;
    return (req == null) ? 0 : req.ushell(cmd,b);
  }

  static class TerminalOutput extends OutputStream {
    public void write(byte[] b,int pos,int len) throws IOException {
      EDXServletRequest req = getRequest();
      if (req == null) return;
      req.getOutputStream().write(b,pos,len);
    }
    public void write(int b) throws IOException {
      EDXServletRequest req = getRequest();
      if (req == null) return;
      req.getOutputStream().write(b);
    }
  }

  private String getStubVersion() {
    PROG prg = u.getActiveTCB().getPROG();
    char ext = prg.getExt();
    if (ext == 0 || (prg.edx.getByte(ext + JAVA_VERS) & 255) == 254)
      return null;
    int i = JAVA_VERS;
    while (prg.edx.getByte(ext + i) != 0) ++i;
    return prg.edx.etoa(ext + JAVA_VERS, i - JAVA_VERS);
  }

  /** The entry point for the interface provided by the
      JAVA, LOADJAVA, and FINDJAVA EDL routines.
      This JAVA interface is built on top of the CALLUSER interface
      provided by Edxuser.
   */
  public static void main(String[] argv) throws Exception {
    if (argv.length == 1) {
      String cmd = argv[0];
      EDX edx = new EDX();
      PROG prg = edx.wheres("JAVA");
      if (prg == null) {
	System.err.println("JAVA not running in EDX");
	return;
      }
      char ext = prg.getExt();
      if (cmd.equals("stop")) {
	if (ext != 0) {
	  System.err.println("Notifying Edxjava to shutdown");
	  prg.edx.post(ext+2,-1);
	}
	else
	  prg.cancel(1);
	return;
      }
      if (cmd.equals("restart")) {
	if (ext != 0) {
	  System.err.println("Requesting Edxjava restart");
	  prg.edx.post(ext+2,1);
	}
	return;
      }
      System.err.println(cmd + " is not a recognized command");
      return;
    }
    Edxuser u = new Edxuser();
    u.commandArgs(argv);
    Edxjava j = new Edxjava(u);
    System.setOut(new PrintStream(new TerminalOutput()));
    try {
      for (;;) {
	Edxuser.UIMsg m = u.getUIMsg(true);
	if (m == null) break;
	if (m.parm == 0 && m.getActiveTCB().equals(m.getLoadingTCB())) {
	  m.uidone(JOK);
	  break;	// Shutdown request from java.e
	}
	try {
	  j.calluser(m);
	}
	catch (Throwable e) {
	  e.printStackTrace();
	  m.uidone(JINTERN);
	}
      }
    }
    catch (EDXException e) {
      // something bad happened, log it and restart
      e.printStackTrace();
      j.userquit(true);
      PROG prg = u.getActiveTCB().getPROG();
      char ext = prg.getExt();
      prg.edx.post(ext+2,3);
      return;
    }
    finally {
      u.dispose();
    }
    j.userquit(false);
  }
}
