/*
 * $Log: Edxmap.java,v $
 * Revision 1.3  2002/01/30 19:25:59  stuart
 * Convert to CVS.
 *
 * Revision 1.2  2001/04/12 23:08:32  stuart
 * Use preparsed format strings
 *
 * Revision 1.1  1998/11/21  00:53:48  stuart
 * Initial revision
 *
 */
package bmsi.edx;

import lava.clib.Stdio;
import lava.clib.stdio.Printf;
import lava.clib.stdio.PrintfFormatString;
import java.util.Vector;

/** Provides high level access to the EDX program load map.
    @author Stuart D. Gathman
    Copyright (C) 1998 Business Management Systems, Inc.
 */
public class Edxmap implements Edxsys, Edxccb, Edxprg, Edxtcb {
  private EDX edx;
  final static PrintfFormatString fmtldpt =
    new PrintfFormatString("%04X  %8.8s  %3d");
  final static PrintfFormatString fmtovr =
    new PrintfFormatString("  %04x %8.8s");
  final static PrintfFormatString fmtuid =
    new PrintfFormatString("%-8d %-20s");
  final static PrintfFormatString fmtccb =
    new PrintfFormatString("%8.8s %-20s");
  final static PrintfFormatString fmtstg =
    new PrintfFormatString(" %3d ");

  /** An extended EDX map entry. */
  public class Proghdr {
    public final char ldpt;		// program load point
    public final short size;		// size in pages
    public final short stg;		// dynamic storage in pages
    public final char ovr;		// overlay address
    public final char ccb;		// PRGCCB for main program
    public final char ptr;		// printer ENQTed, if any
    public final String name;		// program name
    public final String ovname;		// overlay name

    /** Return a 1 line description of this map entry in the format
	used by <code>edxmap</code>.
     */
    public String toString() {
      StringBuffer buf = new StringBuffer(40);
      Stdio.sprintf(buf,fmtldpt, new Object[] {
	new Integer(ldpt),name,new Integer(size)
      });
      String ccbname;
      String uname = "";
      int uid = 0;
      if (ccb != 0) {
	if ((ccb & 1) != 0)
	  ccbname = "N/A";
	else {
	  ccbname = edx.etoa(0,ccb + CCBNAME,8);
	  uid = edx.getWord(0,ccb + CCBUID);
	}
	buf.append(Printf.format(fmtovr,
	  new Object[] { new Integer(ovr),ovname }
	));
	if (stg != 0)
	  buf.append(
	    Printf.format(fmtstg,new Object[] { new Integer(stg) }));
	if ((ptr & 1) != 0 || ptr == ccb || ptr == 0) {
	  buf.append(Printf.format(fmtuid,
	    new Object[] { new Integer(uid),uname }
	  ));
	}
	else {
	  ccbname = edx.etoa(0,ptr + CCBNAME,8);
	  buf.append(Printf.format(fmtccb,
	    new Object[] { ccbname,uname }
	  ));
	}
      }
      return buf.toString();
    }

    /** Construct this extended map entry from a MapEntry. */
    Proghdr(MapEntry map,short key) {
      ldpt = map.ldpt;
      size = map.size;
      switch (map.type) {
      case MapEntry.PROG:
	char tcb = edx.getAddr(key,ldpt + PRGTCB);
	name = edx.etoa(key,ldpt + PRGNAM,8);
	stg = edx.getWord(key,ldpt + PRGLOS);
	ccb = edx.getAddr(key,ldpt + PRGCCB);
	char tptr = findptr(key,ccb,tcb);
	ovr = (char)(edx.getAddr(key,ldpt + PRGPGM) & ~1);
	if (ovr != 0) {
	  if (edx.getWord(key,ovr) == 0x0808) {
	    tcb = edx.getAddr(key,ovr + PRGTCB);
	    tptr = findptr(key,edx.getAddr(key,ovr + PRGCCB),tcb);
	    ovname = edx.etoa(key,ovr + PRGNAM,8);
	  }
	  else
	    ovname = "--";
	}
	else
	  ovname = "";
	ptr = tptr;
	return;
      case MapEntry.FREE:
	name = "  (FREE)";
	break;
      default:
	name = "  (DATA)";
      }
      ovname = "";
      stg = 0;
      ccb = 0;
      ptr = 0;
      ovr = 0;
    }
  }

  public Edxmap(EDX edx) throws EDXException {
    this.edx = edx;
  }

  /** Display all partitions on System.out. */
  public void printAll() {
    int npart = edx.getNumpart();
    for (int key = 0; key < npart; ++key) {
      Proghdr[] hdr = getMap(key);
      if (key > 0)
	System.out.println();
      System.out.print(" " + (key + 1) + " ");
      for (int i = 0; i < hdr.length; ++i) {
	Proghdr h = hdr[i];
	System.out.print("\t");
	System.out.println(h.toString());
      }
    }
  }

  /** Return an array of extended map entries.
     @param key the EDX address key of the partition to map
     @return null if partition does not exist
   */
  public Proghdr[] getMap(int key) {
    MapEntry[] map = edx.getPartitionMap(key);
    if (map == null)
      return null;	// partition not present
    Proghdr[] hdr = new Proghdr[map.length];
    for (int i = 0; i < map.length; ++i)
      hdr[i] = new Proghdr(map[i],(short)key);
    return hdr;
  }

  /** Find the first printer in use by an EDX program.
   */
  private char findptr(short key,char ccb,char tcb1) {
    char tcb = tcb1;
    char prg = edx.getAddr(key,tcb + TCBPLP);
    // search for first TCB with TERMINAL other than default
    do {
      char t = edx.getAddr(key,tcb + TCBCCB);
      if (t != ccb) return t;
      tcb = edx.getAddr(key,tcb +TCBPTR);
    } while (tcb != tcb1);
    // search for CCB pointing at PRGPPLP per extended ENQT semantics
    prg = edx.getAddr(key,prg + PRGPPLP);
    char ccba = edx.getAddr(0,CCBADDR);
    if ((ccba & 1) != 0) return ccb;
    ccba = edx.getAddr(0,ccba);	// first ccb
    while (ccba != 0) {
      if ((ccba & 1) != 0 || edx.getWord(0,ccba + CCBTAG) != 0xFFFF) return ccb;
      if (edx.getAddr(0,ccba + CCBQUIN) == prg
	&& edx.getWord(0,ccba + CCBQUIN + 2) == key)
	return ccba;
      ccba = edx.getAddr(0,ccba + CCBLINK);
    }
    return ccb;
  }

  public static void main(String[] argv) throws Exception {
    String vtab = "/edx";
    if (argv.length > 0)
      vtab = argv[0];
    EDX edx = new EDX(vtab + "/edxvtab.sys");
    Edxmap edxmap = new Edxmap(edx);
    edxmap.printAll();
  }
}
