/*
 * $Log: Edxuser.java,v $
 * Revision 1.4  2002/01/30 19:25:59  stuart
 * Convert to CVS.
 *
 * Revision 1.3  2000/06/20 03:53:19  stuart
 * improved exception handling
 *
 * Revision 1.2  1998/10/31  21:31:38  stuart
 * Provide both active (calling) and loading TCB
 *
 * Revision 1.1  1998/10/27  20:26:23  stuart
 * Initial revision
 *
 */
package bmsi.edx;

import posix.*;

/** The user extension interface to an EDX emulator.  The LOADUSER and
    CALLUSER edl commands can invoke code in a separate process via this
    interface.  After initialization, users of this class should call 
    <code>getUIMsg()</code> in a loop to wait for and process CALLUSER
    commands from edl.

    @author Stuart D. Gathman
 */
public class Edxuser {
  private static boolean pidused = false;

  private static /*final*/ int
    ui_mtype, ui_opword, ui_key, ui_tcb, ui_parm, ui_len;

  static {
    // Compute C struct offsets
    CPtr.Struct s = new CPtr.Struct(0);
    ui_mtype = s.offsetOf(CPtr.CLONG_TYPE);
    ui_opword = s.offsetOf(CPtr.CSHORT_TYPE);
    ui_key = s.offsetOf(CPtr.CSHORT_TYPE);
    ui_tcb = s.offsetOf(CPtr.CSHORT_TYPE);
    ui_parm = s.offsetOf(CPtr.CSHORT_TYPE);
    ui_len = s.size();
  }

  private /*final*/ EDX edx;
  private int uiid = 0;
  private int opword = 0;
  private final Malloc uimsg = new Malloc(ui_len);

  private ActiveTCB active;

  private boolean background;	// true if -B specified
  private String[] argv;	// user command args

  /** Create a CALLUSER interface attached to an EDX interface. */
  public Edxuser(EDX edx) {
    this.edx = edx;
  }

  /** Create a CALLUSER interface attached to the default EDX nucleus. */
  public Edxuser() throws EDXException {
    this.edx = new EDX();
  }

  /** Return TCB access for the task that loaded us via LOADUSER.
     Return null if this interface was not loaded by EDL code.
   */
  public ActiveTCB getActiveTCB() {
    return active;
  }

  /** Register a user interface id, and return a unique 16-bit integer for use
      as the mtype field of the message queue.  This should really be 
      a 32-bit integer, but it is returned as $TCBCO2 to the EDX program.
      The concept is like the IPC key used to obtain an id in the posix API.

      @param uid An application id, or 0 to assign a unique id.
      @return 0 on failure
   */
  public static synchronized int adduid(
    short key,char tcb,short opword,short uid) {
    // FIXME: how to use registry in EDX/MAX ?
    if (pidused) return 0;
    pidused = true;
    return IPC.pid & 0xFFFF;
  }

  /** Release a user interface id assigned with adduid. */
  public static synchronized void deluid(int uid) {
    if (uid == (IPC.pid & 0xFFFF))
      pidused = false;
  }

  /**
  * Initialize user interface for this user ID.
  * The opword parameter is the instruction word used to load this IDL program.
  * Bits 4-7 (mask 0x0700) are used to control the load:
  <pre>
        0x0100 - Do not register this process to be killed on PROGSTOP.
        0x0200 - Load IDL process inforeground (by terminal emulator).
        0x0400 - Reserved.
        0x0800 - Reserved.
  </pre>
  * If the key and tcb parameters are not 0 then load return code (-1 for OK)
  * is returned in TCBCO and the user interface ID is placed in TCBCO2.
  * TCBCO return codes are:
  <pre>
       -1  Successfull load.
        1  ID already in use.	(really 2?)
        2  No more IDs avaiable.(really 1?)
  </pre>
  **/
  public void uiinit(short key,char tcb,short opword,short id)
    throws EDXException {
    if (uiid != 0)
      throw new EDXException("User interface already active");
    uiid = adduid(key,tcb,opword,id);
    // update the TCB if we were loaded by an EDL program.
    if (tcb != 0) {
      active = edx.active(key,tcb);
      active.putWord(tcb + Edxtcb.TCBCO2,uiid);	// return uiid in $TCBCO2
      active.putWord(tcb,(uiid != 0) ? -1 : ((id != 0)? 2 : 1));
      if ((opword & 0x0200) == 0)	// FOREGRND=NO
	active.ready();
    }
    else
      active = null;
    if (uiid == 0)
      throw new EDXException((id != 0)? "UI id in use" : "UI id table full");
  }

  /** A pool of UIMsg instances to avoid excessive garbage collection. */
  UIMsg pool = null;

  /** Return a free UIMsg from the pool - or create a new one if the pool
      is empty. */
  private synchronized UIMsg getUIMsg() {
    UIMsg m = pool;
    if (m != null)
      pool = m.next;
    else
      m = new UIMsg();
    return m;
  }

  /** Each UIMsg represents an active USERCALL instruction from
     the EDX emulator.  When processing is finished, call <code>uidone()</code>
     to resume execution of the EDX task that invoked the USERCALL.
    */
  public class UIMsg {
    UIMsg next;
    short opword;
    short key;
    char tcb;
    char parm;

    /** Return the TCB that invoked us via the EDL CALLUSER instruction.  */
    public final ActiveTCB getActiveTCB() {
      return edx.active(key,tcb);
    }

    /** Return the TCB that loaded this JVM via the EDL LOADUSER instruction. */
    public final ActiveTCB getLoadingTCB() {
      return active;
    }

    /** Return the parameter word (usually the EDX address of a
      parameter block) for this USERCALL.
     */
    public final char getParm() { return parm; }

    /** Return the opcode word for this USERCALL. */
    public final short getOpword() { return opword; }

    /** Finish current USERCALL, and return this instance to the free
	pool.  <b>WARNING:</b> this instance may be recycled at
	any time after you call <code>uidone()</code>.  The public state
	of UIMsg instances is readonly to minimize the damage you
	can do should you attempt to use it again.  However, you
	can still modify the wrong EDX task through getActiveTCB().

	@param code The return code to give back to the EDX program.
     */
    public void uidone(int code) {
      if ((opword & 0x3000) != 0x1000 && tcb != 0) {
	if (code == 0) code = -1;
	edx.putWord(key,tcb,code);
	try {
	  edx.ready(key,tcb);
	}
	catch (Exception x) {
	  System.err.println("UIMsg.uidone: " + x.toString());
	}
      }
      tcb = 0;
      synchronized (Edxuser.this) {
	next = pool;
	pool = this;
      }
    }
  }

  /** Return the next USERCALL message in the queue.

    @param wait set true to wait
    @return the UIMsg or null if PROGSTOP or no messages
   */
  public UIMsg getUIMsg(boolean wait) throws EDXException {
    if (uiid == 0) return null;
    if (edx.usrq == null) return null;
    int flag = wait ? 0 : IPC.IPC_NOWAIT;
    try {
      // only one uimsg object
      synchronized (uimsg) {
	for (;;) {
	  int rc = edx.usrq.recv(uimsg,uiid,flag);
	  if (rc == ui_len) break;
	  // ignore UI messages with wrong size
	}
	short opword = uimsg.getCShort(ui_opword);
	if (opword == 0) {	// got PROGSTOP message
	  dispose();
	  return null;
	}

	UIMsg m = getUIMsg();
	m.opword = opword;
	m.key = uimsg.getCShort(ui_key);
	m.tcb = (char)uimsg.getCShort(ui_tcb);
	m.parm = (char)uimsg.getCShort(ui_parm);
	return m;
      }
    }
    catch (IPCException e) {
      if (e.getErrno() == Errno.EIDRM)
	return null;	// edxnuc was killed
      throw new EDXException(e.getMessage());
    }
  }

  /** Release resources for this USERCALL interface. */
  public void dispose() {
    deluid(uiid);
    uiid = 0;
    uimsg.free();
    active = null;
  }

  public int getuid() { return uiid; }

  /** Parse the standard command args passed by the LOADUSER instruction
      from EDX, and 
      initialize this user interface as per <code>uiinit</code>.

      @return The remaining args for the application.
   */
  public String[] commandArgs(String[] argv) throws EDXException {
    int i = 0;
    short key = 0;
    char tcb = 0;
    for (; i < argv.length; ++i) {
      if (!argv[i].startsWith("-")) break;
      int alen = argv[i].length();
      if (alen > 1) {
	char ac = argv[i].charAt(1);
	switch (ac) {
	case 'K':
	  key = Short.parseShort(
	    (alen > 2) ? argv[i].substring(2) : argv[++i]);
	  continue;
	case 'T':
	  tcb = (char)Integer.parseInt(
	    (alen > 2) ? argv[i].substring(2) : argv[++i],16);
	  continue;
	case 'B':
	  background = true;
	  continue;
	}
      }
      throw new EDXException("LOADUSER undefined command line switch");
    }
    if (i + 2 > argv.length)
      throw new EDXException("LOADUSER interface ID not present");

    short opword = (short)Integer.parseInt(argv[i++],16);
    short uid = Short.parseShort(argv[i++]);
    uiinit(key,tcb,opword,uid);

    String[] a = new String[argv.length - i];
    System.arraycopy(argv,i,a,0,a.length);
    this.argv = a;
    return a;
  }

}
