package bmsi.edx;

import bmsi.util.PrintBuffer;
import java.util.Enumeration;
import java.io.IOException;
import jgl.Deque;
import jgl.HashSet;

/** A class to enumerate the tree of EDX menu options reachable by
    a user.  The enumeration excludes options that are unreachable due
    to the user privilege mask.
 */
public class MenuPath implements Enumeration {
  private static final MenuFile menu = new MenuFile();
  private static final MenuUser user = new MenuUser();
  /** pending output for depth first recursion */
  private final Deque pending = new Deque();	
  /** track menus already seen to prevent infinite recursion */
  private final HashSet menus = new HashSet();
  private int userpriv;

  private MenuPath(MenuOption start,int userpriv) {
    pending.pushBack(start);
    this.userpriv = userpriv;
  }

  public boolean hasMoreElements() { return !pending.isEmpty(); }

  public Object nextElement() {
    MenuOption opt = (MenuOption)pending.popFront();
    if (opt.isMenu() && menus.put(opt.prog) == null)
      addMenu(opt);
    return opt;
  }

  private synchronized void addMenu(MenuOption m) {
    String scrn = m.prog;
    int pos = scrn.indexOf(',');
    if (pos >= 0)
      scrn = scrn.substring(0,pos);
    try {
      synchronized (menu) {
	menu.editKey();
	menu.scrn.value(scrn);
	menu.setMinRange();
	menu.editMaxKey();
	menu.scrn.value(scrn);
	menu.setMaxRange();
	menu.first();
	String path = m.path;
	if (path.length() > 0)
	  path = path + "/";
	while (!menu.eof()) {
	  int menupriv = menu.priv.value();
	  if ((menupriv & userpriv) == menupriv) {
	    MenuOption opt = new MenuOption();
	    opt.path = path + menu.opt.value();
	    opt.prog = menu.prog.value() + "," + menu.vol.value();
	    opt.dir = menu.dir.value();
	    opt.desc = menu.desc.value();
	    pending.pushBack(opt);
	  }
	  menu.next();
	}
      }
    }
    catch (IOException e) { }
  }

  /** Enumerate the menus reachable from a user id.  Each enumeration
    element is a <code>MenuOption</code>.
    @see MenuOption
   */
  public static synchronized Enumeration listMenus(int userid) {
    try {
      menu.open();
      user.open();
      user.editKey();
      user.id.value(userid);
      if (!user.gotoKey()) return null;
      MenuOption start = new MenuOption();
      start.path = "";
      start.prog = user.prog.value() + "," + user.vol.value();
      start.dir = "";
      start.desc = user.name.value();
      return new MenuPath(start,user.priv.value());
    }
    catch (IOException e) {
      System.err.println(e.toString());
    }
    return null;
  }

  /** Invoke <code>listMenus</code> from the command line for a
  set of user ids. */
  public static void main(String[] argv) {
    PrintBuffer txt = new PrintBuffer();
    for (int i = 0; i < argv.length; ++i) {
      Enumeration e = listMenus(Integer.parseInt(argv[i]));
      if (e == null) {
	System.err.println("User " + argv[i] + " not found.");
	continue;
      }
      while (e.hasMoreElements()) {
	MenuOption m = (MenuOption)e.nextElement();
	txt.printAt(0,m.prog);
	txt.printAt(16,m.dir);
	txt.printAt(32,m.desc);
	txt.printAt(54,m.path);
	System.out.println(txt.toString());
      }
    }
  }
}
