/*
 * $Log: PROG.java,v $
 * Revision 1.6  2001/04/12 21:03:07  stuart
 * doc updates
 *
 * Revision 1.5  2001/04/12 20:51:52  stuart
 * rename getBMSOverlayLevel
 * sbrk method
 *
 * Revision 1.4  1999/08/20  02:06:31  stuart
 * provide info about BMS overlay system
 *
 * Revision 1.3  1999/08/10  16:21:37  stuart
 * Hide IPCExceptions
 *
 * Revision 1.2  1998/11/21  00:45:29  stuart
 * get extension block
 * cancel(), make edx,plp public
 *
 */
package bmsi.edx;
import java.util.StringTokenizer;

/** This class provides access to properties of an EDX program.
    @author Stuart D. Gathman
    @version $Revision: 1.6 $
    Copyright (C) 1998 Business Management Systems, Inc.
 */

public class PROG implements Edxprg {
  public final ActiveTCB edx;
  /** The program load point. */
  public final char plp;
  private final char parm;	// parameter count word

  /** Create an instance representing an EDX program.  PROG instances
     are created by the <code>EDX</code> and <code>ActiveTCB</code>
     classes in this package.  FIXME: should take plp only and find 
     and validate the main TCB.  Other TCBs could be removed from the program.
   @param edx	A TCB belong to the EDX program to establish addressing
   @param plp	The EDX program load point
   @see EDX
   @see ActiveTCB
   */
  PROG(ActiveTCB edx,char plp) {
    if ((plp & 0xFF) != 0)
      throw new EDXError("Invalid PLP: "
	+ edx.actkey + '-' + Integer.toHexString(plp));
    this.edx = edx;
    this.plp = plp;
    int dscnt = edx.getWord(plp + PRGDSC);	// number of datasets
    int pgcnt = edx.getWord(plp + PRGPGM);	// number of overlay programs
    parm = (char)(plp + PRGDSN + (dscnt + pgcnt) * Edxdscb.DSCBL);
  }

  /** Cancel this EDX program. This effectively makes the program execute
      PROGSTOP as its next instruction.  NOTE: the EDX virtual
      machine is another process.  The cancel will take effect when
      the EDX VM processes the CANCEL message from its message queue.
      @param code	The PROGSTOP code
   */
  public void cancel(int code) throws EDXException {
    edx.edx.cancel(edx.actkey,plp,(short)code);
  }

  /** Return the name of this EDX program.  
    The name is padded to 8 chars.  Call <code>trim()</code> if needed.
   */
  public String getName() {
    return edx.getText(PRGNAM);
  }

  /** Return the address of the EDX program extension block.
     @return 0 if not present
   */
  public char getExt() {
    return edx.getAddr(plp + PRGEXT);
  }

  /** Return the name of the EDX volume this program was loaded from.
    The name is padded to 6 chars.  Call <code>trim()</code> if needed.
   */
  public String getVol() {
    return edx.etoa(plp + PRGVOL,6);
  }

  /** Return the CCB enqtued when this EDX program was loaded. */
  public final CCB getCCB() {
    return new CCB(edx.edx,edx.getAddr(plp + PRGCCB));
  }

  /** Return the EDX security record for this program. */
  public final EDXLogin getLogin() {
    return EDXLogin.getLogin(edx.edx,edx.getAddr(plp + PRGCCB));
  }

  /** Return an EDX parameter word passed to this EDX program.
      Each EDX program can have up to 99 words of data copied to
      its program header.  These words are known as program parameters.
      The parameters are numbered beginning with 1.  
      @param i the program parameter number to retrieve
      @return the parameter value
   */
  public final short getParm(int i) {
    if (i <= 0 || i > getParmCount())
      throw new ArrayIndexOutOfBoundsException("EDX PROG Parm: " + i);
    return edx.getWord(parm + i * 2);
  }

  /** Return the number of program parameters.
    @since 1.4
    @return the number of program parameter words
   */
  public final short getParmCount() { return edx.getWord(parm); }

  /** Set a program parameter word for this EDX program.
     @since 1.4
     @param i the program parameter to set
     @param val the new value
   */
  public final void putParm(int i,int val) {
    if (i <= 0 || i > getParmCount())
      throw new ArrayIndexOutOfBoundsException("EDX PROG Parm: " + i);
    edx.putWord(parm + i * 2,val);
  }

  /** Return true if this is an EDX overlay program. */
  public final boolean isOverlay() {
    return plp != edx.getAddr(plp + PRGPPLP);
  }

  /** Post the EDX program end event.  This is often used to run in the
      background while the loading program continues.
      @since 1.4
      @param code	the post code for the program end event
      @exception EDXException if there is an IPC error talking to the EDX
      	virtual machine.
   */
  public final void postend(int code) throws EDXException {
    edx.post(edx.getAddr(plp + PRGEECB),code);
  }

  /** Copy parameter bytes to the XPARM area of the BMS overlay stub.
      This is how arguments are passed to BMS overlays.
      @since 1.4
      @param b the argument bytes to pass
      @exception XParmException When this is not a BMS overlay or there
      	is not enough space available in the XPARM area.
   */
  public final void parmPut(byte[] b) {
    if (!isBMSOverlay())
      throw new XParmException("Not a BMS overlay");
    int stk = (char)getParm(4);
    int len = edx.getWord(stk - 2) - edx.getWord(stk - 4) - 2;
    int blength = (b == null) ? 0 : b.length;
    if (blength > len)
      throw new XParmException("XPARM area overflow");
    int stkend = stk + edx.getWord(stk - 2) - 2;
    edx.putWord(stkend,blength);
    if (b != null)
      edx.putBytes(stkend - blength,b);
  }

  /** Copy parameter bytes out of the XPARM area.
      This is how arguments are returned from BMS overlays.
    @since 1.4
    @param b A byte[] to store the retrives parameter bytes.
    @return The number of bytes returned.
   */
  public final int parmGet(byte[] b) {
    if (!isBMSOverlay()) return 0;
    int stk = (char)getParm(4);
    int stkend = stk + edx.getWord(stk - 2);
    int ln = edx.getWord(stkend - 2);
    stkend -= ln;
    if (ln > b.length)
      ln = b.length;
    edx.copyOut(stkend - 2,b,0,ln);
    return ln;
  }

  /** Push an EDX program name on the BMS program stack.  The BMS program
   stack is popped and program executed when the current overlay and
   all pending SHELL calls have finished.
   @since 1.4
   @param prog the NAME,VOL of the edx program to push on the stack
   @exception XParmException if this is not a BMS overlay or there is no
     room in the program stack.
   */
  public final void pushProgram(String prog) {
    if (!isBMSOverlay())
      throw new XParmException("Not a BMS overlay");
    int stk = (char)getParm(4);
    int stkidx = edx.getAddr(stk - 4);
    if (stkidx + 14 > edx.getAddr(stk - 2))
      throw new XParmException("BMS program stack overflow");
    edx.putWord(stk - 4,stkidx + 14);
    StringTokenizer tok = new StringTokenizer(prog,",");
    edx.atoe(stk + stkidx,8,tok.nextToken());
    edx.atoe(stk + stkidx + 8,6,tok.hasMoreTokens()?tok.nextToken():"");
  }

  // GLOBAL PARAMETER AREA IN #MSMAIN ADDRESSED BY $PARM5
  static final short CMPARM   = 0;	// 2 W  PARAMETER WORD
  static final short CMSTGSAV = 2;	// 4 W  STORAGE LOCATION, SIZE
  static final short CMMSOSAV = 6;	// 36    #MSOPEN DSCB SAVE
  static final short CMSTKSAV = 42;	// 36    DISK STACK DSCB SAVE
  static final short CMBTCB   = 78;	// 2 W  DEFAULT DIRECTORY BTCB ADDRESS
  static final short CMCODE   = 80;	// 2 W  PROGRAM RETURN CODE
  static final short CMDSCB   = 82;	// 2 W  DSCB ADDRESS
  static final short CMXPARM  = 84;	// 2 W  XPARM ADDRESS
  static final short CMVER    = 86;	// 2 W  POINTER TO CMPARM

  /** Return true if this EDX program is a BMS overlay.  The BMS overlay
    software uses a compact stub as a main program.  The stub swaps overlay
    programs in and out.  A BMS overlay has an extended API, especially 
    <code>SHELL</code>, which suspends the current overlay while running
    another, then resumes.
    @since 1.4
    @return true if this EDX program is a BMS overlay.
   */
  public final boolean isBMSOverlay() {
    if (!isOverlay() || getParmCount() != 5) return false;
    int cm = (char)getParm(5);
    return cm == edx.getAddr(cm + CMVER);
  }

  /** Return the disk stack offset of this BMSOverlay.  Since EDX tasks
  	can be suspended and overlaid by the BMS overlay system, an
	active task can have the same TCB address as a suspended task.
	Java objects that are associated with a particular EDX task,
	should therefore check this property.
    @return offset in stack dataset, or 0 if this is not a BMS overlay.
   */
  public int getBMSOverlayLevel() {
    if (!isOverlay() || getParmCount() != 5) return 0;
    int cm = (char)getParm(5);
    if (cm != edx.getAddr(cm + CMVER)) return 0;
    return edx.getDword(cm + CMSTKSAV + Edxdscb.DSCBNEX - Edxdscb.DSCBORN);
  }

  /** Allocate EDL dynamic storage.  This is done by incrementing the
     dynamic storage address and decrementing the storage length of 
     the primary EDX program.  The storage is allocated in 256 byte
     pages.  These pages will be independent of pages allocated by
     EDL code using <code>MEMGET</code> and <code>MEMPUT</code>.
     It is necessary for Java code to allocate EDX memory to
     use some EDX features, for example the menu support
     in the BMS 4978 client.  While we try to allocate the memory
     as atomically as possible, it is <u>not</u> safe to 
     call <code>sbrk</code> concurrently with allocation by EDL code.
     EDL allocates dynamic storage in <code>MEMGET</code> when its
     free list is exhausted.  This typically occurs only during
     initialization, before any Java code is invoked, but you should
     keep this in mind.
     @since 1.5
     @param size The number of bytes to allocated, rounded to a multiple
     	of 256.
     @return The EDX address of the allocated storage, or 0 if no 
     	dynamic storage is available.
   */
  int sbrk(int size) {
    int pplp = edx.getAddr(plp + PRGPPLP);	// primary program load point
    int pages = (size + 255) >> 8;
    int stg = edx.getAddr(pplp + PRGSTG);
    int los = edx.getAddr(pplp + PRGLOS);
    if (los < pages || los > 255) return 0;
    edx.putWord(pplp + PRGSTG,stg + pages * 256);
    edx.putWord(pplp + PRGLOS,los - pages);
    return stg;
  }

}
