package bmsi.edx;

/** Conversions between Series/1 floating-point format and Java.
  Series/1 format is slightly weird.
  There's a sign bit where you'd expect it to be, a 7-bit exponent (which they
  call a "characteristic"), and a 24-bit fraction.  The wierd part is that
  the exponent is base 16, a sneaky way of extending the range of their numbers
  in exchange for a sort of statistical loss of precision.  Normalization on
  the Series/1 just means making sure that the top four bits of the fraction --
  the top hex "digit" if you will -- are non-zero; therefore there may be as
  many as three zero bits at the top of the fraction, which are wasted if you're
  thinking in base-2 terms.  Then again, there may not.
<p>
  Note that there's a bit of lost precision when we convert back to Series/1
  format after an operation.  At the moment we're not rounding, we're just 
  dropping the bottom end of the IEEE fraction into the bit bucket.  Hope 
  nobody notices.  
<p>
  Pnuts test script:
<pre>
f = class bmsi.edx.S1Float
a = [ 0.5, 25.0, Double::POSITIVE_INFINITY, Math::PI, -0.25 ]
a = a + [ -0.1, Double::NEGATIVE_INFINITY, -1234567890.1 ]
foreach d (double[0] + a) {
  dw = f::doubleToS1FloatBits(d)
  System::err.print("" + d + " -> " + Integer::toHexString(dw))
  System::err.println(" -> " + f::s1FloatBitsToDouble(dw))
  lw = f::doubleToS1ExtBits(d)
  System::err.print("" + d + " -> " + Long::toHexString(lw))
  System::err.println(" -> " + f::s1ExtBitsToDouble(lw))
}
</pre>

Results:
<pre>
0.5 -> 40800000 -> 0.5
0.5 -> 4080000000000000 -> 0.5
25.0 -> 42190000 -> 25.0
25.0 -> 4219000000000000 -> 25.0
Infinity -> 7fffffff -> Infinity
Infinity -> 7fffffffffffffff -> Infinity
3.141592653589793 -> 413243f6 -> 3.141592025756836
3.141592653589793 -> 413243f6a8885a30 -> 3.141592653589793
-0.25 -> c0400000 -> -0.25
-0.25 -> c040000000000000 -> -0.25
-0.1 -> c0199999 -> -0.09999996423721313
-0.1 -> c01999999999999a -> -0.1
-Infinity -> ffffffff -> -Infinity
-Infinity -> ffffffffffffffff -> -Infinity
-1.2345678901E9 -> c8499602 -> -1.23456768E9
-1.2345678901E9 -> c8499602d2199998 -> -1.2345678901E9
</pre>

  @author Stuart D. Gathman
  Converted from the float.c module of EDX/MAX.
 */

public class S1Float {

  public final static int MAX_FLOAT = 0x7FFFFFFF;
  public final static long MAX_EXT = 0x7FFFFFFFFFFFFFFFL;


  private static double ldexp(long mant,int exp) {
    /* normalize the mantissa so we don't lose range in the exponent. */
    while ((mant & 0xFFF0000000000000L) == 0) {
      mant <<= 1;
      --exp;
    }
    return Double.longBitsToDouble(
      mant & 0x000FFFFFFFFFFFFFL | ((long)(exp + 51 + 0x400) << 52));
  }

  /** Convert Series/1 floating point bits to a double.  The result is 
      a double to preserve the larger range of the S/1 format. */
  public static double s1FloatBitsToDouble(int s1f) {
    if (s1f == -1)
      return Double.NEGATIVE_INFINITY;
    if (s1f == Integer.MAX_VALUE)
      return Double.POSITIVE_INFINITY;
    /* Grab the S1 exponent (7-bit), move it to the bottom of a word,
	un-bias it (64), and multiply it by four to convert from base-16 to
	base-2.  To save time, we combine the right-shift 24 (move to bottom)
	and the left-shift 2 (multiply by 4) and adjust the bias accordingly. */
    int exponent = (s1f & 0x7f000000) >> 22;
    int fraction = s1f & 0xffffff;
    if (fraction == 0 || exponent == 0)
      return 0.0;

    exponent -= 0x100 + 52;	// unbias the exponent

    long mant = (long)fraction << 28;
    if ((s1f & 0x80000000) != 0)
      return -ldexp(mant,exponent);
    return ldexp(mant,exponent);
  }

  /** Convert Series/1 extended floating point bits to a double.  The result is 
      a double to preserve the larger range of the S/1 format. */
  public static double s1ExtBitsToDouble(long s1x) {
    if (s1x == -1L)
      return Double.NEGATIVE_INFINITY;
    if (s1x == Long.MAX_VALUE)
      return Double.POSITIVE_INFINITY;
    /* Grab the 7-bit exponent, shift and unbias... */
    int exponent = (int)(s1x >> 54) & (0x7f << 2);
    long fraction = s1x & 0x00FFFFFFFFFFFFFFL;
    if (exponent == 0 || fraction == 0L)
      return 0.0;
    while ((fraction & 0xFFE0000000000000L) != 0) {
      fraction >>= 1;
      ++exponent;
    }
    exponent -= 0x100 + 56;

    if ((s1x & 0x8000000000000000L) != 0)
      return -ldexp(fraction,exponent);
    return ldexp(fraction,exponent);
  }

  /** Convert a double to Series/1 floating point bits. */
  public static int doubleToS1FloatBits(double f) {
    int sign = (f < 0) ? 0x80000000 : 0;
    if (Double.isInfinite(f))
      return MAX_FLOAT | sign;
    long bits = Double.doubleToLongBits(f);
    // extract and rebias the exponent
    int exponent = ((int)(bits >> 52) & 0x7FF) - 0x400 + 0x100 + 2;
    int fraction = ((int)(bits >> 29) & 0xFFFFFF | 0x800000);
    if (fraction == 0)
      return 0;

    /*  Normalize the fraction in the S1 sense, which is to say make the
	exponent be divisible by 4 cuz it's gonna be base-16 shortly.  */
    int rem = exponent & 3;
    if (rem > 0) {
      rem = 4 - rem;
      fraction >>= rem;
      exponent += rem;
    }

    /*  Build the S1 number.  The S1 exponent is 7-bit, biased 64, and base 
	16.  What we've got right now is 11-bit, unbiased, and base 2.  We 
	just happen to know for sure that the bottom two bits of the exponent 
	are zero, so we're gonna very sneakily bias it by 4*64, and then shift 
	it two too few places to save a shift.  */
    return sign | (exponent << 22) | fraction;
  }

  /** Convert a double to Series/1 extended format. */
  public static long doubleToS1ExtBits(double f) {
    long bits = Double.doubleToLongBits(f);
    if (Double.isInfinite(f))
      return MAX_EXT | (bits & 0x8000000000000000L);
    long fraction = (bits & 0x000FFFFFFFFFFFFFL | 0x0010000000000000L) << 3;
    // extract and rebias the exponent
    if (fraction == 0L) 
      return 0L;
    /* Grab the 11-bit exponent and rebias it... */
    int exponent = ((int)(bits >> 52) & 0x7FF) - 0x400 + 0x100 + 2;

    /*  Normalize the fraction in the S1 sense, which is to say make the
	exponent be divisible by 4 cuz it's gonna be base-16 shortly.  */
    int rem = exponent & 3;
    if (rem > 0) {
      rem = 4 - rem;
      fraction >>= rem;
      exponent += rem;
    }

    return (bits & 0x8000000000000000L) | ((long)exponent << 54) | fraction;
  }
}
