/*
 * $Log: SelectPrinter.java,v $
 * Revision 1.5  2002/01/29 21:36:49  stuart
 * Support printer setup escapes
 *
 * Revision 1.4  2001/04/12 20:41:56  stuart
 * move to bmsi.edx package
 *
 * Revision 1.3  2000/11/28  18:18:08  stuart
 * Set terminal environment
 *
 * Revision 1.2  2000/11/16 00:34:51  stuart
 * Handle piped printers and CRNL property
 *
 * Revision 1.1  2000/02/23 18:39:38  stuart
 * Initial revision
 *
 */
package bmsi.edx;

import java.io.*;
import bmsi.util.StanzaFile;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/** A class that maps EDX logical printer names to OutputStreams.
    @version $Id: SelectPrinter.java,v 1.5 2002/01/29 21:36:49 stuart Exp $
    @author Stuart D. Gathman
 */
public class SelectPrinter {

  private StanzaFile termxref;
  private TerminalPrinter termptr;
  private Hashtable termenv = new Hashtable();

  public SelectPrinter() {
    termptr = new TerminalPrinter(true);
  }

  /** Get the full list of logical printer names defined by the Termxref file.
   */
  public String [] getPrinterList() {
    StanzaFile termxref = this.termxref;
    if (termxref == null)
      return new String[0];
    return termxref.getStanzaNames();
  }

  /** Get the list of logical printer names for a terminal location.
   * The BTAS/X file /edx/TERMINAL.PRINTER lists the printers displayed
   * for each terminal location.  If this file is not found or the location
   * is not found in the file, then all printers in the Termxref file
   * are listed.
   */
  public synchronized String [] getPrinterList(String termname) {
  getlist:
    try {
      StanzaFile termxref = getTermXref();
      if (termxref == null)
	break getlist;
      termptr.open();
      termptr.editKey();
      termptr.term.value(termname);
      termptr.ptr.value("");
      termptr.gotoNearest();
      Vector list = new Vector();
      while (!termptr.eof() && termptr.term.value().equals(termname)) {
	String ptr = termptr.ptr.value();
	if (termxref.getStanzaValue(ptr) == null)
	  continue;	// printer not in termxref
	list.addElement(ptr);
	termptr.next();
      }
      if (list.size() == 0)
	break getlist;
      String[] a = new String[list.size()];
      list.copyInto(a);
      return a;
    }
    catch (IOException e) { }
    finally {
      termptr.close();
    }
    return getPrinterList();
  }

  /** Get the list of terminal location names defined by TERMINAL.PRINTER.
   */
  public synchronized String [] getTerminalList() {
    Vector list = new Vector();
    try {
      termptr.open();
      String lastterm = new String();
      for (termptr.first(); !termptr.eof(); termptr.next()) {
	String term = termptr.term.value();
	if (!term.equals(lastterm)) {
	  list.addElement(termptr.term.value());
	  lastterm = term;
	}
      }
    }
    catch (IOException e) { }
    finally {
      termptr.close();
    }
    String[] a = new String[list.size()];
    list.copyInto(a);
    return a;
  }

  /** Get a PrintStream corresponding to a printer in the Termxref file.
    @return null if the printer is not found in the Termxref file.
    @deprecated use getOutputStream
   */
  public PrintStream getPrinter(String ptr) throws IOException {
    return new PrintStream(getOutputStream(ptr));
  }

  private StanzaFile getTermXref() throws IOException {
    StanzaFile termxref = this.termxref;
    if (termxref != null) {
      termxref.refresh();
      return termxref;
    }
    do {
      setTermxref(getDefaultTermxref());
      termxref = this.termxref;
    } while (termxref == null);
    return termxref;
  }

  /** Add a CR to each NL not followed by a CR. */
  static class CRNLOutputStream extends FilterOutputStream {
    private int last = -1;

    public CRNLOutputStream(OutputStream out) { super(out); }
    public void write(int b) throws IOException {
      if (last != '\r' && b == '\n' && last != '\n')
	out.write('\r');
      out.write(last = b);
    }
    public void write(byte[] buf,int pos,int len) throws IOException {
      if (len <= 0) return;
      if (last != '\r' && last != '\n' && buf[pos] == '\n')
	out.write('\r');
      int end = pos + len;
      for (int i = pos + 1; i < end; ++i) {
	if (buf[i] == '\n') {
	  int c = buf[i - 1];
	  if (c != '\r' && c != '\n') {
	    out.write(buf,pos,i - pos);
	    out.write('\r');
	    pos = i;
	  }
	}
      }
      out.write(buf,pos,end - pos);
      last = buf[end - 1];
    }
  }

  /** Set an environment variable for external printer processes.
    @param key	the environment variable name
    @param val	the value
   */
  public void setTermEnv(String key,String val) {
    if (val == null)
      termenv.remove(key);
    else
      termenv.put(key,val);
  }

  /** Return the terminal environment as an array of String with the
     form <code>NAME=VALUE</code>, such as used by Runtime.exec().
   */
  public String[] getTermEnv() {
    synchronized (termenv) {
      String[] a = new String[termenv.size()];
      Enumeration e = termenv.keys();
      for (int i = 0; i < a.length; ++i) {
        String k = e.nextElement().toString();
	a[i] = k + '=' + termenv.get(k);
      }
      return a;
    }
  }

  public byte[] getEscape(String ptr,String prop) throws IOException {
    String esc = getTermXref().getStanzaProperty(ptr,prop);
    if (esc == null) return null;
    return getEscape(esc);
  }

  /** Decode String representation of a termxref escape sequence.
   @param val	The String representation
   @return the binary escape sequence
   */
  static public byte[] getEscape(String val) {
    ByteArrayOutputStream p = new ByteArrayOutputStream(val.length());
    int c;
    int s = 0;
    while (s < val.length()) {
      c = val.charAt(s++);
      char cs = (s < val.length()) ? val.charAt(s) : 0;
      switch (c) {
      case '^':	/* a control char */
	if (cs == '?')
	  c = 127;
	else if (cs != 0)
	  c = cs & 0x1F;
	++s;
	break;
      case '\\':
	if (cs != 0) {
	  ++s;
	  switch (c = cs) {
	  case 'E':		c = 27; break;
	  case 'n':		c = '\n'; break;
	  case 'b':		c = '\b'; break;
	  case 'r':		c = '\r'; break;
	  case 's':		c = ' '; break;
	  case 'x': 	
	    c = 0;
	    while (s < val.length()) {
	      int d = Character.digit(val.charAt(s),16);
	      if (d < 0) break;
	      c *= 16;
	      c += d;
	      ++s;
	    }
	    break;
	  case '0':
	    c = 0;
	    while (s < val.length()) {
	      int d = Character.digit(val.charAt(s),8);
	      if (d < 0) break;
	      c *= 8;
	      c += d;
	      ++s;
	    }
	    break;
	  }
	}
      }
      p.write(c);
    }
    return p.toByteArray();
  }

  /** Get an OutputStream corresponding to a printer in the Termxref file.
    @return null if the printer is not found in the Termxref file.
   */
  public OutputStream getOutputStream(String ptr) throws IOException {
    StanzaFile termxref = getTermXref();
    String cmd = termxref.getStanzaValue(ptr);
    if (cmd == null || cmd.length() < 1)
      return null;
    OutputStream out = null;
    if (cmd.charAt(0) != '|')
      out = new FileOutputStream(cmd);
    else {
      Process proc = Runtime.getRuntime().exec(
	new String[] { "sh", "-c", cmd.substring(1) },
	getTermEnv()
      );
      out = proc.getOutputStream();
    }
    String cropt = termxref.getStanzaProperty(ptr,"CRNL");
    boolean crnl = (cropt == null) || cropt.toUpperCase().startsWith("Y");
    if (crnl)
      out = new CRNLOutputStream(out);
    return out;
  }

  /** Set the name of the file that maps printer names to devices or command
   * pipelines. By default, the bmsi.edxdir property is used as the directory
   * and 'termxref.sys' as the filename.  The bmsi.edxdir property
   * defaults to "/edx".
   */
  public void setTermxref(String fname) throws IOException {
    termxref = new StanzaFile(fname);
  }

  /** Get the default terminal location.  This comes from
   the EDXTERM terminal environment variable, or from the
   <code>bmsi.term</code> system property if EDXTERM is not set.
   */
  public String getDefaultTerminal() {
    Object t = termenv.get("EDXTERM");
    if (t != null) return t.toString();
    return System.getProperty("bmsi.term","$SYSLOG");
  }

  /** Get the default termxref file name. */
  public static String getDefaultTermxref() {
    String edxdir = System.getProperty("bmsi.edxdir","/edx");
    return edxdir + "/termxref.sys";
  }

  /** list the printers for terminal locations. 
   */
  public static void main(String[] args) {
    try {
    SelectPrinter sel = new SelectPrinter();
    String [] list = null;
    if (args.length < 1) {
      sel.setTermxref(sel.getDefaultTermxref());
      list = sel.getPrinterList(sel.getDefaultTerminal());
    }
    else {
      int i = 0;
      if (args[0].length() > 0 && args[0].charAt(0) == '/')
	sel.setTermxref(args[i++]);
      if (i < args.length)
	list = sel.getPrinterList(args[i]);
      else
	list = sel.getPrinterList();
    }
    for (int i = 0; i < list.length; ++i)
      System.out.println(list[i]);
    }
    catch (IOException e) {
      System.err.println(e);
    }
  }
}
