/*
 * $Log: DBTable.java,v $
 * Revision 1.11  2002/08/06 17:40:10  stuart
 * Use class name to reference static final constants.
 *
 * Revision 1.10  2002/03/20 01:31:22  stuart
 * RecordEdit.show()
 *
 * Revision 1.9  2002/02/12 20:53:36  stuart
 * Record edit framework.
 * getRow, getCol in FSPanel.
 * Don't post invalid record in DBTable.
 *
 * Revision 1.8  2002/02/06 04:29:44  stuart
 * Fix removeBuffers when deleting first line in grid.
 * Consume pfkeys.
 * Advance to next field on ENTER.
 *
 * Revision 1.7  2002/01/29 21:47:18  stuart
 * Fix scrolling.  Eat PFKeys handled.
 *
 * Revision 1.6  2000/10/20 21:46:06  stuart
 * try to get scrolling to work more like EDL
 *
 * Revision 1.5  2000/06/17 21:56:13  stuart
 * Make NEXT/PREV (F2/F3) work more like FCB
 *
 * Revision 1.4  2000/06/17  20:42:02  stuart
 * set the display length of fields when painting.
 *
 * Revision 1.3  2000/05/26 20:12:30  stuart
 * Handle deleted rows
 *
 */
package bmsi.edx.fs;

import java.io.IOException;
import bmsi.db.*;

/** A database table control for EDX.  This was a major function of
    the FCB statement in the BMS EDL library.  The location of the
    table on the screen is specified with a beginning and ending 
    field and a number of lines.  The beginning and ending field
    define the first line, which is copied to give the full line count.
    @author Stuart D. Gathman
    Copyright (C) 1999 Business Management Systems, Inc.
 */
public class DBTable extends FSGrid implements DataListener,BufferListener {
  private Field[] fldmap;
  private DataScroller ds;
  private int firstRecord;
  private int activeRecord;
  private boolean modified;
  private final boolean[] deleted;
  /** Next field to posCursor when tabbing out of table. */
  public int next;

  /** Create a DBTable. The first row between the first and last fields
    will be duplicated to create additional lines.
   @param pan	the panel containing the fields for this DBTable
   @param ds	the Dataset with the records displayed/edited by this table
   @param beg	the first field index of the first row of the table
   @param end	the last field index of the first row of the table
   @param cnt	the total number of screen lines for the table
   */
  public DBTable(FSPanel pan,DataScroller ds,int beg,int end,int cnt) {
    super(pan,beg,end,cnt);
    this.ds = ds;
    ds.addDataListener(this);
    ds.addBufferListener(this);
    deleted = new boolean[cnt];
  }

  /** Set the data fields displayed in the columns controlled by this DBTable.
    By default, the first n fields of the DataScroller are displayed - where
    n is the number of columns in the table.
   */
  public void setFlds(Field[] f) {
    fldmap = f;
  }

  private Field getField(int col) {
    if (fldmap == null) {
      fldmap = new Field[ds.fieldCount()];
      for (int i = 0; i < fldmap.length; ++i)
	fldmap[i] = ds.fields(i);
    }
    try {
      Field f = fldmap[col];
      if (f != null)
	f.displayLength(panel.getFieldLen(beg + col));
      return f;
    }
    catch (IndexOutOfBoundsException x) { }
    return null;
  }

  /** Return data for a FSGrid cell as a String from the corresponding
    database field.
    @param row	the grid row
    @param col	the grid col
    @return the current field value as a String, or null if no value available
   */
  public String getText(int row,int col) {
    Field f = getField(col);
    if (f != null)
      return ds.getText(firstRecord + row,f);
    return null;
  }

  /** Set the value of the database field corresponding to a FSGrid 
     cell as a String.
    @param row	the grid row
    @param col	the grid col
    @param s	the value to set as a String
   */
  public void setText(int row,int col,String s) {
    Field f = getField(col);
    if (f != null && !f.isReadonly() && f.getDataBuffer().isEditing())
	f.asString(s);
  }

  private int lastDist;
  /** Called by the DataScoller when the dataset cursor has been moved.  */
  public void datasetScrolled(DataScroller ds,int distance) {
    int rows = getRecordCount();
    int arec = ds.getActiveRecord();
    //System.err.print("activerec="+arec+" firstRecord="+firstRecord+
    //	" dist="+distance+" rows="+rows);
    if (distance == 0 || Math.abs(distance) >= rows) {
      // redisplay all records
      if (arec < rows)
	firstRecord = 0;
      else {
	firstRecord = arec - rows + 1;
      }
      paint();
      lastDist = 0;
    }
    else {
      lastDist = distance;
      int rowsel = arec - firstRecord;
      //System.err.print(" rowsel="+rowsel);
      if (rowsel < 0) {
	firstRecord = arec;
	//System.err.print(" ins="+(-rowsel));
	insertRows(0,-rowsel);
      }
      else if (rowsel >= rows) {
	int cnt = rowsel - rows + 1;
	firstRecord += cnt;
	//System.err.print(" del="+cnt);
	deleteRows(0,cnt);
      }
    }
    //System.err.println(" newfirst="+firstRecord);
    activeRecord = arec;
    setSelection(arec - firstRecord,colsel);
  }

  /** Called by the DataScoller to delete lines from the display.  */
  public void removeBuffers(int idx,int cnt) {
    //System.err.println("removeBuffers(idx="+idx+",cnt="+cnt+")");
    //try { throw new Exception("Trace"); }
    //catch (Exception x) { x.printStackTrace(); }
    if (cnt > 0) {
      // removing cnt buffers starting at idx
      if (idx >= firstRecord) {
	deleteRows(idx - firstRecord,cnt);
      }
      if (idx < firstRecord)
	firstRecord -= cnt;
      //deleteRows(0,cnt);
    }
    else {
      // inserting -cnt buffers before idx
      if (idx < firstRecord)
	firstRecord -= cnt;
      else
	insertRows(idx - firstRecord,-cnt);
    }
  }

  /** Called by the DataScoller to discover the number of lines displayed. */
  public final int getRecordCount() { return getRows(); }

  /** Called by the DataScoller when the editing mode changes. */
  public void editingChanged(DataScroller ds) { }

  /** Called by the DataScoller when a field in the record changes.
    @param d	the DataScroller 
    @param f	the field that changed, or null if all or unknown
   */
  public void recordChanged(DataScroller d,Field f) {
    int row = ds.getActiveRecord() - firstRecord;
    if (f == null)
      paintRow(row);
    else
      for (int i = beg; i <= end; ++i) {
	Field fld = getField(i-beg);
	if (f == fld) {
	  if (d == ds)
	    setLine(row);
	  panel.putField(i,f.asString(),false);
	  return;
	}
      }
  }

  /** Called by the DataScoller to update the record from screen data.  */
  public void updateData(DataScroller ds) {
    if (!modified) return;
    int row = ds.getActiveRecord() - firstRecord;
    setLine(row);
    for (int i = beg; i <= end; ++i) {
      setText(rowsel,i-beg,panel.getField(i));
    }
    modified = false;
    if (deleted[row]) {
      deleted[row] = false;
      try {
	ds.delete();
	return;
      }
      catch (IOException x) { }
    }
  }

  /** Mark the selected line as deleted. */
  public void zeroBuf() {
    //System.err.println("delete row " + rowsel);
    deleted[rowsel] = true;
  }

  /** Called by the DataScoller when the Dataset is opened or closed. */
  public void activeChanged(DataScroller ds) { }

  /** Position the DataScroller to the selected line.  Do not 
    update the screen - the screen data is what we want to do
    something with in this case.
   */
  public void rdline() throws IOException {
    //System.err.println("rdline row=" + rowsel);
    int active = firstRecord + rowsel;
    int dist = active - ds.getActiveRecord();
    //System.err.println("dist="+dist+" active="+active);
    if (dist > 0)
      ds.next(dist);
    else if (dist < 0)
      ds.prev(-dist);
  }

  /** Handle standard PFKEYs pressed when the cursor is in this table.  */
  public void processPfkey(PFKeyEvent e) {
    super.processPfkey(e);
    if (!e.consumed)
      try {
	switch (e.pfkey) {
	case PFKeyEvent.PF_ENTER: {
	  e.consume();
	  int n = next;
	  //if (n == 0 || contains(panel.getRow(),panel.getCol()))
	  //  n = 0;
	  if (ds.isEditing()) {
	    if (!ds.post())
	      ds.cancel();
	  }
	  else if (n > 0) {
	    panel.posCursor(n);
	    return;
	  }
	  break;
	}
	case PFKeyEvent.PF_PREV: {
	  e.consume();
	  int r = ds.getActiveRecord() - firstRecord;
	  int n = getRows() - 1;
	  ds.prev(n + r);
	  if (ds.bof())
	    ds.first();
	  setSelection(0,0);
	  rdline();
	  return;
	}
	case PFKeyEvent.PF_NEXT: {
	  e.consume();
	  int r = ds.getActiveRecord() - firstRecord;
	  int n = getRows() - 1;
	  if (n > r)
	    ds.next(n - r);	// goto end of page
	  if (!ds.eof()) {
	    ds.next(n);		// goto end of next page
	    if (ds.eof()) {
	      firstRecord += n - lastDist;
	      paint();
	    }
	  }
	  setSelection(0,0);
	  rdline();
	  return;
	}
	case PFKeyEvent.PF_FIND:
	  if (modified && ds instanceof Dataset) {
	    try {
	      Dataset ds = (Dataset)this.ds;
	      ds.editKey();
	      ds.gotoNearest();	// search by key
	    }
	    catch (Exception x) {
	      x.printStackTrace();
	    }
	    break;
	  }
	  if (rowsel > 0) {
	    rdline();
	    firstRecord += rowsel;	// put line at top
	    ds.next(getRows() - 1);	// ensure screen within buffers
	    //ds.prev(getRows());
	    paint();
	    setSelection(0,0);
	    rdline();
	    return;
	  }
	  break;
	case PFKeyEvent.PF_INSERT:		// insert 
	  if (ds.readonly()) break;
	  e.consume();
	  rdline();
	  if (ds.isEditing()) 
	    if (!ds.post())
	      break;
	  ds.insert();
	  break;
	case PFKeyEvent.PF_DELETE:		// delete
	  if (ds.readonly()) break;
	  e.consume();
	  rdline();
	  ds.delete();
	  break;
	case PFKeyEvent.FIELD_MODIFIED:
	  if (!ds.readonly()) {
	    rdline();
	    if (!ds.isEditing() && ds.recordValid())
	      ds.edit();
	  }
	  modified = true;
	  break;
	case PFKeyEvent.FIELD_ENTER:
	  rdline();
	  return;
	case PFKeyEvent.FIELD_EXIT:
	  return;
	}
	setSelection(ds.getActiveRecord() - firstRecord,0);
      }
      catch (IOException x) {
        panel.centerLine(23,x.toString());
      }
  }

}
