/*
 * $Log: FSPanel.java,v $
 * Revision 1.11  2002/08/06 16:56:31  stuart
 * New FSPanel ctor.
 *
 * Revision 1.1  2002/07/01 22:09:23  stuart
 * Initial revision
 *
 * Revision 1.10  2002/02/12 20:53:36  stuart
 * Record edit framework.
 * getRow, getCol in FSPanel.
 * Don't post invalid record in DBTable.
 *
 * Revision 1.9  2002/01/29 21:45:43  stuart
 * FIELD_ENTER event
 *
 * Revision 1.8  2002/01/24 01:40:44  stuart
 * Fix typo
 *
 * Revision 1.7  2002/01/24  01:28:08  stuart
 * Track min,max rows for lindup
 *
 * Revision 1.6  2001/12/03  21:35:30  stuart
 * expand tabs in loadImage, trap invalid field exceptions
 *
 * Revision 1.5  2001/03/23 19:31:59  stuart
 * support ASCII screen images
 *
 * Revision 1.4  2000/11/10 16:46:22  stuart
 * scrolling
 * restore cursor when reactivated
 * make choice case insensitive unless choices are mixed case
 *
 * Revision 1.3  2000/06/27 16:04:02  stuart
 * bug in choice()
 *
 * Revision 1.2  2000/05/30 16:07:03  stuart
 * improved modified field handling
 *
 * Revision 1.1  2000/02/04  19:17:40  stuart
 * Initial revision
 *
 */
package bmsi.edx.fs;

import bmsi.xopen.IsamFile;
import bmsi.xopen.Xopen;
import bmsi.xopen.FType;
import bmsi.edx.EDX4978;
import bmsi.edx.EDXError;
import bmsi.edx.CCBOwner;
import bmsi.edx.CCB;
import bmsi.util.Str;
import java.util.Vector;
import java.io.Reader;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;

/** An EDX application screen.  Data input fields can be defined which 
   are referenced by field index.  A panel can be initialized from a
   BTAS file.  This roughly corresponds to what the EDX FS library provided.
   A panel must be painted onto an EDX4978 to be displayed.  The panel
   saves a reference to the EDX4978 when painted and keeps it updated.
   An application might have several panels, painting them in turn as
   they become active.

   @see bmsi.edx.EDX4978

   @author Stuart D. Gathman
   Copyright (C) 1999 Business Management Systems, Inc.
 */
public class FSPanel extends FSComp implements CCBOwner {
  /** Field table elements. */
  private static final int
    ROW = 0, COL = 1, LEN = 2, MIN = 3, MAX = 4;
  private static final int COLS = 80;
  private static final int ROWS = 24;
  private static final byte PROT_SPACE = (byte)0xA0;

  private byte[] buf = new byte[ROWS * COLS];
  private byte[] newbuf = new byte[ROWS * COLS];
  /** Low level fields defined in screen image. */
  private final Vector flds = new Vector();
  /** Components on this screen. */
  private final Vector comps = new Vector();
  private EDX4978 scrn;
  private boolean hasFocus;
  private boolean hasScreen;
  private boolean fldexit;

  public void addComponent(FSComp comp) {
    comps.removeElement(comp);
    comps.addElement(comp);
  }

  public void removeComponent(FSComp comp) {
    comps.removeElement(comp);
  }

  /** Add a new input field. */
  public void addField(int row,int col,int len) {
    byte[] f = new byte[3];
    f[ROW] = (byte)row;
    f[COL] = (byte)col;
    f[LEN] = (byte)len;
    flds.addElement(f);
  }

  /** Position the cursor at the beginning of an input field. */
  public void posCursor(int fld) {
    byte[] f = (byte[])flds.elementAt(fld);
    scrn.setCursor(f[ROW],f[COL]);
  }

  /** Position the cursor at an arbitrary location. */
  public void posCursor(int row,int col) {
    scrn.setCursor(row,col);
  }

  /** Display a string in a data field.
    @param fld	The screen field to write to
    @param s	The String to write
    @param prot	True to protect the field
  */
  public void putField(int fld,String s,boolean prot) {
    byte[] f = (byte[])flds.elementAt(fld);
    //scrn.setCursor(f[0],f[1]);
    int pos = scrn.getPos(f[ROW],f[COL]);
    scrn.write(pos,s,f[LEN],prot);
    scrn.read(pos,buf,pos,f[LEN]);
  }

  /** Find changes to a field.
    @param fld the field to scan
    @return the offset of the first change, or -1 if no changes
   */
  public int isModified(int fld) {
    byte[] f = (byte[])flds.elementAt(fld);
    int pos = scrn.getPos(f[ROW],f[COL]);
    int len = f[LEN];
    byte[] b = new byte[len];
    scrn.read(pos,b,0,len);
    for (int i = 0; i < b.length; ++i)
      if (b[i] != buf[pos++]) return i;
    return -1;
  }

  /** Return screen data for an input field as a String. */
  public String getField(int fld) {
    try {
      byte[] f = (byte[])flds.elementAt(fld);
      int row = f[ROW];
      int col = f[COL];
      if (hasScreen) return scrn.input(row,col);
      int cnt = 80 - col;
      byte[] b = new byte[cnt];
      int i = scrn.getPos(row,col);
      System.arraycopy(buf,i,b,0,cnt);
      return scrn.input(b);
    }
    catch (ArrayIndexOutOfBoundsException x) {
      throw new IllegalArgumentException("fld = " + fld);
    }
  }

  /** Return the size in chars of an input field. */
  public int getFieldLen(int fld) {
    try {
      byte[] f = (byte[])flds.elementAt(fld);
      return f[LEN];
    }
    catch (ArrayIndexOutOfBoundsException x) {
      throw new IllegalArgumentException("fld = " + fld);
    }
  }

  public int getFieldMin(int fld) {
    byte[] f = (byte[])flds.elementAt(fld);
    return f.length < 4 ? f[ROW] : f[MIN];
  }

  public int getFieldRow(int fld) {
    try {
      byte[] f = (byte[])flds.elementAt(fld);
      return f[ROW];
    }
    catch (ArrayIndexOutOfBoundsException x) {
      throw new IllegalArgumentException("fld = " + fld);
    }
  }

  public void setFieldRow(int fld,int row) {
    byte[] f = (byte[])flds.elementAt(fld);
    f[ROW] = (byte)row;
  }

  public int getFieldCol(int fld) {
    byte[] f = (byte[])flds.elementAt(fld);
    return f[COL];
  }

  /** Return the number of data input fields for this panel. */
  public int getFieldCount() { return flds.size(); }

  /** Locate the field containing a screen position.
    @return the field index, or -1 if none
   */
  public int getFieldAt(int row,int col) {
    int n = flds.size();
    for (int i = 0; i < n; ++i) {
      byte[] f = (byte[])flds.elementAt(i);
      if (f[COL] <= col && col < f[COL] + f[LEN]) {
        if (f[ROW] == row) return i;
	if (f.length >= 5 && f[MIN] <= row && row <= f[MAX]) return i;
      }
    }
    return -1;
  }

  public FSComp getComponentAt(int row,int col) {
    int n = comps.size();
    while (n-- > 0) {
      FSComp c = (FSComp)comps.elementAt(n);
      if (c.contains(row,col))
        return c;
    }
    return this;
  }

  /** Duplicate a row of fields onto multiple lines to form a table.
    @param beg	The first field.
    @param end	The last field.
    @param cnt	The number of lines.
    @return The first screen line
   */
  int lindup(int beg,int end,int cnt) {
    byte[] f = (byte[])flds.elementAt(beg);
    int row = f[ROW];
    int begpos = row * COLS + f[COL];
    f = (byte[])flds.elementAt(end);
    if (f[ROW] != row)
      /* We could actually handle this, but punt for now. */
      throw new IllegalArgumentException("Lindup for multiple lines");
    int len = f[ROW] * COLS + f[COL] + f[LEN] - begpos;
    int pos = begpos;
    for (int i = beg; i <= end; ++i) {
      // save min,max row in f[3],f[4]
      byte[] of = (byte[])flds.elementAt(i);
      byte[] nf = new byte[5];
      int min = of[ROW];
      nf[ROW] = (byte)min;
      nf[COL] = of[COL];
      nf[LEN] = of[LEN];
      nf[MIN] = (byte)min;
      nf[MAX] = (byte)(min + cnt - 1);
      flds.setElementAt(nf,i);
    }
    for (int i = 1; i < cnt; ++i) {
      pos += COLS;
      System.arraycopy(buf,begpos,buf,pos,len);
      if (scrn != null)
        scrn.write(pos,buf,pos,len);
    }
    return row;
  }

  /** Scroll a table formed by lindup.  The top line must be set
    with setFieldRow for <code>beg</code> and <code>end</code>.
    @param beg The first field.
    @param end The last field.
    @param cnt The number of lines to scroll
    @param dist The distance to scroll, < 0 to copy up
   */
  void scroll(int beg,int end,int cnt,int dist) {
    byte[] f = (byte[])flds.elementAt(beg);
    int row = f[ROW];
    int begpos = row * COLS + f[COL];
    f = (byte[])flds.elementAt(end);
    int len = f[ROW] * COLS + f[COL] + f[LEN] - begpos;
    if (dist < 0) {
      int pos = begpos + dist * COLS;
      for (int i = 0; i < cnt; ++i) {
	System.arraycopy(buf,begpos,buf,pos,len);
	//System.err.println(new String(buf,pos,len));
	if (scrn != null)
	  scrn.write(pos,buf,pos,len);
	pos += COLS;
	begpos += COLS;
      }
    }
    else {
      begpos += cnt * COLS;
      int pos = begpos + dist * COLS;
      for (int i = 0; i < cnt; ++i) {
        pos -= COLS;
	begpos -= COLS;
	System.arraycopy(buf,begpos,buf,pos,len);
	if (scrn != null)
	  scrn.write(pos,buf,pos,len);
      }
    }
  }

  /** Initialize an FSPanel to protected blanks. */
  public FSPanel() {
    super(null);
    fill(buf,0,buf.length,PROT_SPACE);
  }

  public static void fill(byte[] b,int pos,int len,byte val) {
    if (len < 4) {
      switch (len) {
      case 3:
	b[pos++] = val;
      case 2:
	b[pos++] = val;
      case 1:
	b[pos] = val;
      }
      return;
    }
    b[pos] = val;
    b[pos+1] = val;
    int i = 2;
    for (; i + i < len; i += i)
      System.arraycopy( b, pos, b, pos + i, i);
    System.arraycopy( b, pos, b, pos + i, len - i);
  }

  /** Initialize an FSPanel from a BTAS file created by
    <code>edximg</code> or <code>#IMAGE</code>. */
  public FSPanel(String imagefile) throws EDXError {
    super(null);
    try {
      IsamFile bt = new IsamFile(imagefile,Xopen.INPUT);
      byte[][] ba = new byte[ROWS][COLS + 2];
      int n = bt.read(ba,Xopen.FIRST);
      bt.close();
      convertImage(ba,n);
    }
    catch (FileNotFoundException e) {
      throw new EDXError(imagefile + ": not found");
    }
    catch (IOException x) {
      throw new EDXError(x.toString());
    }
  }

  public FSPanel(InputStream in,char nulch) throws EDXError {
    super(null);
    try { loadImage(new InputStreamReader(in),nulch); }
    catch (IOException x) { throw new EDXError(x.toString()); }
  }

  /** Convert array of BTAS records from EBCDIC image file to EDX4978 image. */
  private void convertImage(byte[][] ba,int n)
    throws UnsupportedEncodingException {
    FType conv = FType.getConverter("Cp037");
    flds.removeAllElements();
    int row = 0;
    int pos = 0;
    byte[] txt = new byte[COLS];
    for (int i = 0; i < n; ++i) {
      byte[] b = ba[i];
      int r = FType.ldshort(b,0);
      //System.err.println("row =" + r + " b.length=" + b.length);
      if (r > row) {
	int cnt = (r - row) * COLS;
	fill(buf,pos,cnt,PROT_SPACE);	// protected spaces
	pos += cnt;
      }
      System.arraycopy(b,2,txt,0,COLS);

      /* Java and EDX/UNX disagree over codes for EBCDIC square brackets. */
      for (int j = 0; j < txt.length; ++j) {
        switch (txt[j]) {
	case (byte)0xAD: txt[j] = (byte)0xBA; break;
	case (byte)0xBD: txt[j] = (byte)0xBB; break;
	}
      }

      char[] ca = conv.convertAll(txt);
      addLine(pos,ca,COLS,(char)0);
      pos += COLS;
      if (false) {
	int flen = 0;
	for (int j = 0; j < COLS; ++j) {
	  int c = ca[j];
	  if (c == 0)
	    ++flen;
	  else if (flen > 0) {
	    addField(r,j-flen,flen);
	    flen = 0;
	  }
	  buf[pos++] = (c == 0) ? 0 : (byte)(c|0x80);
	}
	if (flen > 0)
	  addField(r,COLS-flen,flen);
      }
      row = r + 1;
    }
    fill(buf,pos,buf.length - row * COLS,PROT_SPACE);
  }

  private void addLine(int pos,char[] ca,int len,char nulch) {
    int flen = 0;
    int r = pos / COLS;
    for (int j = 0; j < len; ++j) {
      int c = ca[j];
      if (c == nulch)
	++flen;
      else if (flen > 0) {
	addField(r,j-flen,flen);
	flen = 0;
      }
      buf[pos++] = (c == nulch) ? 0 : (byte)(c|0x80);
    }
    if (flen > 0)
      addField(r,len-flen,flen);
  }

  public void loadImage(Reader r,char nulch) throws IOException {
    BufferedReader rdr = (r instanceof BufferedReader)
      ? (BufferedReader)r : new BufferedReader(r);
    char [] ca = new char[COLS];
    int pos = 0;
    for (;;) {
      String s = rdr.readLine();
      if (s == null) break;
      int len = Str.expandTabs(s,ca);
      addLine(pos,ca,len,'_');
      pos += COLS;
    }
  }

  /** Display this FSPanel on an EDX4978, and make the EDX4978 the
    current output device.
    @param sp	The EDX4978 device to display on, or null to stop displaying.
   */
  public void paint(EDX4978 sp) {
    scrn = sp;
    if (sp != null)
      paint();
  }

  private int crow, ccol;

  public int getRow() { return crow; }
  public int getCol() { return ccol; }

  public void deactivated(CCB ccb) {
    hasScreen = false;
    if (scrn == ccb) {
      crow = scrn.getRow();
      ccol = scrn.getCol();
      //System.err.println("saved row = " + crow + " col = " + ccol);
    }
  }

  public void activated(CCB ccb) {
    hasScreen = true;
  }

  public void paint() {
    if (scrn != null) {
      scrn.setOwner(this);
      scrn.write(0,buf,0,buf.length);
      scrn.setCursor(crow,ccol);
      int n = comps.size();
      while (n-- > 0) {
	FSComp c = (FSComp)comps.elementAt(n);
	c.paint();
      }
    }
  }

  /** Process pfkey events for this panel until focus is lost.
    @see transferFocus
   */
  public void fprocess() { fprocess(true); }

  public int pfkey;

  /** Process pfkey events for this panel.
    @param loop	true to loop until focus is lost, otherwise process
    	exactly one pfkey
   */
  public void fprocess(boolean loop) {
    if (scrn == null) return;
    if (scrn.getRow() == 0 && scrn.getCol() == 0 && flds.size() > 0)
      posCursor(0);
    hasFocus = loop;
    do {
      int row = scrn.getRow();	// save cursor before pfkey
      int col = scrn.getCol();
      pfkey = scrn.waitkey(fldexit);
      crow = scrn.getRow();	// new cursor location
      ccol = scrn.getCol();
      if (pfkey == PFKeyEvent.FIELD_EXIT) {
	//System.err.println("("+row+','+col+") -> ("+crow+','+ccol+')');
        int fld = getFieldAt(row,col);
	if (fld >= 0) {
	  // FIXME: isModified not checking proper line of lindup
	  int c = isModified(fld);
	  if (c >= 0)
	    sendPFKey(PFKeyEvent.FIELD_MODIFIED,row,col + c);
	}
	/* Send FIELD_EXIT to location we are leaving. */
        sendPFKey(pfkey,row,col);
	sendPFKey(PFKeyEvent.FIELD_ENTER,crow,ccol);
	continue;
      }
      findChanges();
      sendPFKey(pfkey,crow,ccol);
    } while (hasFocus);
  }

  /** Set field exit mode.  In field exit mode, a pseudo pfkey is generated
  	when the user moves between physical fields.
    @param f true to generate <code>PFKeyEvent.FIELD_EXIT</code> when
    	leaving a physical field.
   */
  public void setFldExit(boolean f) { fldexit = f; }

  private void sendPFKey(int pfkey,int row,int col) {
    //System.err.println("PF"+pfkey+'@'+row+','+col);
    FSComp c = getComponentAt(row,col);
    PFKeyEvent pfevt = new PFKeyEvent(c,pfkey,row,col);
    c.processPfkey(pfevt);
    if (!pfevt.consumed)
      processPfkey(pfevt);
  }

  /** Read new screen buffer and scan for changes.  */
  private void findChanges() {
    scrn.read(0,newbuf,0,newbuf.length);	// new screen contents
    // FIXME: shouldn't we swap buffers before scanning?
    int pos = 0;
    for (int r = 0; r < ROWS; ++r) {
      int flen = 0;
      boolean changed = false;
      for (int c = 0; c < COLS; ++c,++pos) {
        int b = newbuf[pos];
	if (b >= 0) {	// found unprotected field
	  ++flen;
	  if (b != buf[pos])
	    /* FIXME: eliminate changed flag by scanning for flen and
	       sending PFKey here.
	     */
	    changed = true;
	}
	else if (flen > 0) {
	  if (changed) {
	    sendPFKey(PFKeyEvent.FIELD_MODIFIED,r,c - flen);
	    changed = false;
	  }
	  flen = 0;
	}
      }
      if (flen > 0 && changed)
        sendPFKey(PFKeyEvent.FIELD_MODIFIED,r,COLS - flen);
    }
    byte[] t = buf;
    buf = newbuf;
    newbuf = t;
  }

  /** Return true if a row and col is within this component.
   */
  public boolean contains(int row,int col) {
    return row >= 0 && row < ROWS && col >= 0 && col < COLS;
  }

  /** Release control of the focus. Currently, this simply returns
    from fprocess().  
    @see fprocess
   */
  public void transferFocus() {
    hasFocus = false;
  }

  public void processPfkey(PFKeyEvent e) {
    super.processPfkey(e);
    if (!e.consumed) switch (e.pfkey) {
    case 1: transferFocus();
    }
  }

  /** Write a String as centered protected text on a line.  This could
    be used as a status or prompt line.
    @param line	The line to overwrite
    @param s	The String to display
   */
  public void centerLine(int line,String s) {
    if (scrn != null)
      scrn.write(line,0,centerPad(s,80,' '),80,true);
  }

  /** Write a prompt and return a 1 character response.  The first
    response char is used as the default.
    @param line	The line to overwrite
    @param msg	The prompt to display
    @param opts	A String with valid response chars.
   */
  public char choice(int line,String msg,String opts) {
    centerLine(line,msg);
    int scol = 40 + (msg.length() + 1)/2;
    String ans = opts.substring(0,1);
    scrn.write(line,scol,ans,1,false);
    for (;;) {
      scrn.setCursor(line,scol);
      scrn.waitkey(true);
      ans = scrn.input(line,scol);
      if (opts.indexOf(ans) >= 0) break;
      ans = ans.toUpperCase();
      if (opts.indexOf(ans) >= 0) break;
    }
    return ans.charAt(0);
  }

  private static StringBuffer sbuf = new StringBuffer(80);

  /** Pad a message to center it.
    @param s	the message to center
    @param len	the desired message length
    @param pad	the char to use for padding
   */
  public static synchronized String centerPad(String s,int len,char pad) {
    int slen = s.length();
    int prepad = (len - slen) / 2;
    int postpad = len - prepad - slen;
    sbuf.setLength(0);
    while (prepad-- > 0) sbuf.append(pad);
    sbuf.append(s);
    while (postpad-- > 0) sbuf.append(pad);
    return sbuf.toString();
  }

}
